""" This is a RDKIT wrapper code that generate reaction network"""

from rdkit import Chem
from rdkit.Chem.AllChem import ReactionFromSmarts
from rdkit.Chem.rdchem import GetPeriodicTable
from rdkit.Chem.rdchem import PeriodicTable
from itertools import product as itpd
from .. RINGParser import Read


def GenerateRxnNet(initial_reactant, reaction_rules):
    """
    Generates reaction network following the algorithm from
    (Ind. Eng. Chem. Res. 2010, 49 (21), 10459-10470)

    Arguments:
    - initial_reactant:     can be smiles or mol
    - reaction_rules:       can be smarts string or rdkit.Chem.rdChemReactions.
                            ChemicalReaction object

    Return:
    - list of reaction intermediates

    Example:
    Ethane C-H scission and C-C scission
    A = generate_rxn_net('CC',['[C:1][H:2]>>[C:1].[H:2]',
                         '[C:1][C:2]>>[C:1].[C:2]'])
    for product in A:
        print(Chem.MolToSmiles(product))

    Tip:
    For dehydrogenation, you must number carbon and hydrogen i.e.
    [C][H]>>[C].[H]             (x)
    [C:1][H:1]>>[C:1].[H:2]     (o)
    Same goes for changing bond order
    [C][C]>>[C]=[C]             (x)
    [C:1][C:2]>>[C:1]=[C:2]     (o)
    Aromatization and Kekulize has some problem with several species. So, we
    don't set aromatization for sanitize, and try kekulize.

    TODO:
    - record reactions as well (make it an option as it's expensive)

    """

    # set-up reactants
    if not isinstance(initial_reactant, list):
        initial_reactant = [initial_reactant]
    if isinstance(initial_reactant[0], str):
        for i in range(0, len(initial_reactant)):
            initial_reactant[i] = Chem.MolFromSmiles(initial_reactant[i],
                                                     sanitize=False)
            # sanitize everything except aromatization set
            _sanitize_except_aromatization(initial_reactant[i])
    # Treatment necessary for radicals
    # (https://github.com/rdkit/rdkit/issues/69)
    for i in range(0, len(initial_reactant)):
        # print Chem.MolToSmiles(initial_reactant[i])
        initial_reactant[i] = Chem.AddHs(initial_reactant[i])
        # sanitize everything except aromatization set
        _sanitize_except_aromatization(initial_reactant[i])
        # Chem.SanitizeMol(initial_reactant[i])
        # Chem.Kekulize(initial_reactant[i])
        for atoms in initial_reactant[i].GetAtoms():
            atoms.SetNoImplicit(True)
        Chem.AssignRadicals(initial_reactant[i])

    # set up reactions
    if not isinstance(reaction_rules, list):
        reaction_rules = [reaction_rules]
    if isinstance(reaction_rules[0], str):
        for i in range(0, len(reaction_rules)):
            try:
                reaction_rules[i] = Read(reaction_rules[i])
            except Exception:
                reaction_rules[i] = ReactionFromSmarts(reaction_rules[i])

    # generator main algorithm
    unprocessed = initial_reactant
    processed = []
    while unprocessed:
        # Pop a molecule and put it in a processed list
        reactant0 = unprocessed[0]
        processed.insert(0, unprocessed[0])
        del unprocessed[0]
        # go through all reactions
        for reaction_rule in reaction_rules:
            # set up reactant list.
            # Generate combinatorial product list of reactants if reaction
            # requires several reactants
            reactant_list = itpd([list(range(1, len(processed)))],
                                 repeat=reaction_rule.
                                 GetNumReactantTemplates()-1)
            # go through each set of reactants
            for reactant_indexes in reactant_list:
                # Reaction
                # Make the reactant mol tuple (Runreactants only accept tuple)
                reactants = (reactant0,)
                for reactant_index in reactant_indexes:
                    reactants += (processed[reactant_index],)
                # React
                ele_reactions = reaction_rule.RunReactants(reactants)

                # Record reactions (TODO)

                # Pre-processing products
                # Go through reactiosn and make a single list of products
                products = []
                for ele_reaction in ele_reactions:
                    for mol in ele_reaction:
                        products.append(mol)
                # Treatment necessary for radicals
                # (https://github.com/rdkit/rdkit/issues/69)
                for mol in products:
                    for atoms in mol.GetAtoms():
                        atoms.SetNoImplicit(True)
                        atoms.UpdatePropertyCache(strict=False)
                    Chem.AssignRadicals(mol)
                    # Remove molecule with atoms with over valence
                for i in range(len(products)-1, -1, -1):
                    for atoms in products[i].GetAtoms():
                        if PeriodicTable.GetDefaultValence(GetPeriodicTable(),
                                                           atoms.GetAtomicNum()
                                                           ) < \
                             atoms.GetTotalValence():
                            del products[i]
                            break
                # remove duplicates
                # TODO. This removes also species with different charges
                for i in range(len(products)-1, -1, -1):
                    for j in range(0, i):
                        if products[i].GetNumAtoms() ==\
                            products[j].GetNumAtoms() and \
                            products[i].GetNumAtoms() ==\
                                len(products[i].
                                    GetSubstructMatch(products[j])):

                            del products[i]
                            break
                # update unprocessed molecule list
                # check for duplicate and append to unprocessed_list if missing
                for mol1 in products:
                    inthelist = 0
                    for mol2 in processed:
                        # first check the nubmer of atoms and then
                        # look for substructure match
                        if mol1.GetNumAtoms() == mol2.GetNumAtoms() and \
                            mol1.GetNumAtoms() == len(mol1.GetSubstructMatch
                                                      (mol2)):
                            # if it's in processed list, break
                            inthelist = 1
                            break
                    # not in the processed list. append to unprocessed
                    if inthelist == 0:
                        unprocessed.insert(0, mol1)
    # Prettify
    for i in range(0, len(processed)):
        # print Chem.MolToSmiles(processed[i])
        processed[i] = Chem.RemoveHs(processed[i], sanitize=False)
        _sanitize_except_aromatization(processed[i])
        # print Chem.MolToSmiles(processed[i])
    return processed


def _sanitize_except_aromatization(mol):
    try:
        Chem.SanitizeMol(mol)
    except Exception:
        Chem.SanitizeMol(mol, sanitizeOps=Chem.rdmolops.SanitizeFlags.
                         SANITIZE_ADJUSTHS)
        Chem.SanitizeMol(mol, sanitizeOps=Chem.rdmolops.SanitizeFlags.
                         SANITIZE_CLEANUP)
        Chem.SanitizeMol(mol, sanitizeOps=Chem.rdmolops.SanitizeFlags.
                         SANITIZE_CLEANUPCHIRALITY)
        Chem.SanitizeMol(mol, sanitizeOps=Chem.rdmolops.SanitizeFlags.
                         SANITIZE_FINDRADICALS)
        try:
            Chem.SanitizeMol(mol, sanitizeOps=Chem.rdmolops.SanitizeFlags.
                             SANITIZE_KEKULIZE)
        except Exception:
            pass
        Chem.SanitizeMol(mol, sanitizeOps=Chem.rdmolops.SanitizeFlags.
                         SANITIZE_PROPERTIES)
        Chem.SanitizeMol(mol, sanitizeOps=Chem.rdmolops.SanitizeFlags.
                         SANITIZE_SETCONJUGATION)
        Chem.SanitizeMol(mol, sanitizeOps=Chem.rdmolops.SanitizeFlags.
                         SANITIZE_SETHYBRIDIZATION)
        Chem.SanitizeMol(mol, sanitizeOps=Chem.rdmolops.SanitizeFlags.
                         SANITIZE_SYMMRINGS)



"""
=======================
YAML-formatted data I/O
=======================

This module handles input from YAML formatted data.  Because this code is in
bad need of cleanup and because in the long run, I (Stephen Edie) hope to
replace the input mechanism with something cleaner and faster (not involving
YAML), the documentation for this module is limited at this time.


---------
Reference
---------

.. autodoc puts stuff extracted from docstrings here.
"""


from .yaml_io import *

__all__ = yaml_io.__all__[:]

del yaml_io



import re
from ..Error import UnitsParseError
from .db import units_db


__all__ = ['eval_expr']


class UnitsParser(object):
    # Regular expression for tokenizing units declarations.
    # TODO: Support evaluation of numbers with e/E exponents.
    tokenize_re = re.compile(r'-?[.\d]+|[a-zA-Z]+|.')

    def __init__(self, expr, debug=False):
        self.tokens = [tok for tok in re.findall(self.tokenize_re, expr)
                       if not tok.isspace()]
        # print 'tokens:', self.tokens
        self.idx = 0
        self.depth = 0
        self.debug = debug

    def isnumber(self, what):
        try:
            float(what)
            return True
        except ValueError:
            return False

    # These two are used to output debugging info.
    def enter(self, what):
        if self.debug:
            print('%s%r {' % (' '*(self.depth*4), what))
            self.depth += 1

    def leave(self, what):
        if self.debug:
            self.depth -= 1
            print('%s} %r' % (' '*(self.depth*4), what))

    def peek(self):
        if self.idx == len(self.tokens):
            return None
        return self.tokens[self.idx]

    def take(self):
        val = self.peek()
        if val is not None:
            self.idx += 1
        return val

    def fork(self):
        return self.idx

    def join(self, old_idx):
        self.idx = old_idx

    def parse(self):
        self.enter('[root]')
        result = None
        try:
            result = self.parse_expr()
            next = self.peek()
            if next is not None:
                result = None
                raise UnitsParseError("Unexpected token '%s'\n" % next)
        finally:
            self.leave(result)
        return result

    def parse_expr(self):
        self.enter('expr')
        result = None
        try:
            result = ('expr', self.parse_factor())
            next = self.peek()
            while True:
                if next is None:
                    # result = None
                    break
                elif next in '*/':
                    self.take()
                    result = ('expr', result, next, self.parse_factor())
                    next = self.peek()
                else:
                    old_state = self.fork()
                    try:
                        result = ('expr', result, '*', self.parse_factor())
                        next = self.peek()
                    except UnitsParseError:
                        self.join(old_state)
                        break
        finally:
            self.leave(result)
        return result

    def parse_factor(self):
        self.enter('factor')
        result = None
        try:
            left = self.parse_base()
            next = self.peek()
            if next == '^':
                self.take()
                result = ('factor', left, '^', self.parse_number())
            else:
                result = ('factor', left)

        finally:
            self.leave(result)
        return result

    def parse_base(self):
        self.enter('base')
        result = None
        try:
            next = self.peek()
            if next is None:
                result = None
                raise UnitsParseError(
                    "Expected name or number, got end of input instead")
            elif next == '(':
                self.take()
                expr = self.parse_expr()
                if self.take() != ')':
                    raise UnitsParseError(
                        "Expected closing parenthesis, got '%s' instead"
                        % next)
                result = ('base', expr)
            elif self.isnumber(next):
                result = ('base', self.parse_number())
            else:
                result = ('base', self.parse_name())

        finally:
            self.leave(result)
        return result

    def parse_name(self):
        self.enter('name')
        result = None
        try:
            next = self.take()
            if next is None:
                raise UnitsParseError(
                    "Expected name instead of end of input")
            if not next.isalpha():
                raise UnitsParseError(
                    "Expected name, got '%s' instead" % next)
            result = ('name', next)

        finally:
            self.leave(result)
        return result

    def parse_number(self):
        self.enter('number')
        result = None
        try:
            next = self.take()
            if next == '(':
                next = self.take()
                closing = self.take()
                if closing != ')':
                    raise UnitsParseError("Expected closing parenthesis, "
                                          "got '%s' instead" % closing)
            if next is None:
                raise UnitsParseError(
                    "Expected number instead of end of input")
            if self.isnumber(next):
                if '.' in next:
                    result = ('number', float(next))
                else:
                    result = ('number', int(next))
            else:
                raise UnitsParseError(
                    "Expected number, got '%s' instead" % next)

        finally:
            self.leave(result)
        return result


def parse(expr, *args, **kwargs):
    """Parse units expression (expr) and return abstract syntax tree."""
    return UnitsParser(expr, *args, **kwargs).parse()


def eval_subtree(tree):
    """
    Evaluate subtree of abstract syntax tree (obtained from
    :func:`parser.parse`).
    """
    if tree[0] == 'expr':
        # Subtree represents an *expression*.
        if len(tree) > 2:
            if tree[2] == '*':
                # Multiply two *factors*.
                return eval_subtree(tree[1])*eval_subtree(tree[3])
            elif tree[2] == '/':
                # Divide two *factors*.
                return eval_subtree(tree[1])/eval_subtree(tree[3])
        else:
            return eval_subtree(tree[1])

    elif tree[0] == 'factor':
        # Subtree represents a *factor* in an expression.
        if len(tree) > 2:
            if tree[2] == '^':
                return eval_subtree(tree[1])**eval_subtree(tree[3])
        else:
            return eval_subtree(tree[1])

    elif tree[0] == 'base':
        # Subtree represents the *base* of an exponentiation.
        return eval_subtree(tree[1])

    elif tree[0] == 'name':
        # Subtree represents *name* of a unit; lookup unit and return.
        return units_db.lookup(tree[1])

    elif tree[0] == 'number':
        # Subtree represents a *number*.
        return tree[1]

    # This shouldn't happen...
    assert False


def eval_expr(expr):
    """
    Evaluate string expression of physical quantity or units.
    """
    ast = parse(expr)
    return eval_subtree(ast)



# -*- coding: utf-8 -*-
from .. Error import RINGError


class Reader(object):
    """
    Reader reads the parsed RING input, and returns the RDkit wrapper objects
    in pgradd.RDkitWrapper.

    Attributes
    ----------
    ast : abstract syntax tree obtrained from parser
    """
    def __init__(self, ast):
        # ast = Abstract Syntax Tree
        self.ast = ast

    def ReadRINGInput(self, tree):
        # Check the type of input
        assert tree[0][0].name in ('Fragment',
                                   'ReactionRule',
                                   'EnumerationQuery')
        # if fragment, molquery is returned
        if tree[0][0].name == 'Fragment':
            from . MolQueryRead import MolQueryReader
            self.type = 'MolQuery'
            return MolQueryReader(tree[0][1:]).Read()
        # if reaction rule, reacitonquery is returned
        elif tree[0][0].name == 'ReactionRule':
            from . ReactionQueryRead import ReactionQueryReader
            self.type = 'ReactionQuery'
            return ReactionQueryReader(tree[0][1:]).Read()
        # TODO enumeration query
        elif tree[0][0].name == 'EnumerationQuery':
            raise NotImplementedError('Coming soon')

    def Read(self):
        # Root tree reading. Check if the input is RINGinput
        assert self.ast[0].name == "RINGInput"
        return self.ReadRINGInput(self.ast[1:])


def Read(text, strict=False):
    """
    Return MolQuery, ReactionQuery, or ReactionNetworkEnumerationQuery by
    interpretting RING input string.

    Parameters
    ----------
    text : string
        Specify string describing chemical structure, elementary reaction, or
        reaction network enumeration rules in RING notation.
    strict : boolean, optional
        If True, then disable use of syntactic extensions such as support for
        "radical electrons".

    Returns
    -------
    Returns RDkit wrapped queries that extends RDkit's functionality:
    * MolQuery if fragment is given in string
    * ReactionQuery if reaction rule is given in string
    * ReactionNetworkEnumerationQuery if enumeration query is given in string

    Raises
    ------
    msr.error.RINGSyntaxError
        If `text` does not conform to RING syntax.
    msr.error.RINGReaderError
        If `text` is invalid RING for non-syntactic reasons.
    """
    from . import Parser
    try:
        return Reader(Parser.parse(text)).Read()
    except RINGError as exc:
        raise exc



from collections.abc import Mapping, Sequence

from .. import Units

from .common import string, InputDataError
# from .schema import *
# from .lib_interface import *


# All classes named _*_schema define builtin schemas for SchemaRepository
# objects.


class any_loader(object):
    def __init__(self, repo):
        pass

    def __call__(self, name, value, context):
        return value


class qty_loader(object):
    def __init__(self, repo, kind=None):
        self.kind = kind

    def __call__(self, name, value, context):
        if value is None:
            return None
        if 'units' in context:
            kind_units = context['units'].get(self.kind)
        else:
            kind_units = None

        qty = Units.eval_qty(value)
        if not isinstance(qty, Units.Quantity):
            if kind_units is not None:
                return Units.with_units(qty, kind_units)
            else:
                raise InputDataError(
                    'Cannot determine units of quantity: %r' % qty)
        return qty


class bool_loader(object):
    def __init__(self, repo):
        pass

    def __call__(self, name, value, context):
        if value is None:
            return None
        try:
            return bool(value)
        except ValueError:
            raise InputDataError('Invalid int: %r' % value)


class float_loader(object):
    def __init__(self, repo):
        pass

    def __call__(self, name, value, context):
        if value is None:
            return None
        try:
            return float(value)
        except ValueError:
            raise InputDataError('Invalid float: %r' % value)


class int_loader(object):
    def __init__(self, repo):
        pass

    def __call__(self, name, value, context):
        if value is None:
            return None
        try:
            return int(value)
        except ValueError:
            raise InputDataError('Invalid int: %r' % value)


class number_loader(object):
    def __init__(self, repo):
        pass

    def __call__(self, name, value, context):
        if value is None:
            return None
        if isinstance(value, int) or isinstance(value, float):
            return value
        try:
            return int(value)
        except ValueError:
            try:
                return float(value)
            except ValueError:
                raise InputDataError('Invalid int: %r' % value)


class string_loader(object):
    def __init__(self, repo):
        pass

    def __call__(self, name, value, context):
        if value is None:
            return None
        try:
            return string(value)
        except ValueError:
            raise InputDataError('Invalid string: %r' % value)


class enum_loader(object):
    # TODO: more elegant -- support type coercion of values to
    # allowable types.
    def __init__(self, repo, values):
        # if values is None:
        #    raise InputDataError("In schema: enum type description is missing"
        #        "required attribute 'values'")
        # elif not isinstance(values, Sequence):
        if not isinstance(values, Sequence):
            raise InputDataError("In schema: enum type description attribute",
                                 "'values' must be a sequence type.  Instead",
                                 "got %r" % values)
        self.values = set(values)

    def __call__(self, name, value, context):
        if value not in self.values:
            raise InputDataError('Enumerated value %r is not one of %r'
                                 % (value, self.values))
        return value


class tuple_loader(object):
    def __init__(self, repo, item_types):
        if not isinstance(item_types, Sequence):
            raise InputDataError(
                'In tuple schema: item_types must be a sequence')
        self.loaders = [repo.make_loader(item_type)
                        for item_type in item_types]

    def __call__(self, name, values, context):
        if (not isinstance(values, Sequence)
                or len(values) != len(self.loaders)):
            raise InputDataError(
                'Expected sequence of %d values, got %r instead'
                % (len(self.loaders), values))
        items = []
        for (i, loader) in enumerate(self.loaders):
            items.append(loader(None, values[i], context))
        return items


class list_loader(object):
    def __init__(self, repo, item_type=None):
        self.loader = repo.make_loader(item_type)

    def __call__(self, name, input_values, context):
        if not isinstance(input_values, Sequence):
            raise InputDataError(
                'Expected sequence, got %r instead' % input_values)
        return [self.loader(string(i), item, context)
                for (i, item) in enumerate(input_values)]


class array_loader(object):
    def __init__(self, repo, shape=None, dtype='d'):
        if not isinstance(shape, Sequence):
            shape = (shape,)
        self.shape = shape
        self.dtype = dtype
        # self.loader = repo.make_loader(item_type)

    def __call__(self, name, input_values, context):
        import numpy
        if not isinstance(input_values, Sequence):
            raise InputDataError(
                'Expected sequence, got %r instead' % input_values)
        ar = numpy.array(input_values, dtype=self.dtype)
        if self.shape is not None and ar.shape != self.shape:
            raise InputDataError("Shape of array %r does not conform to "
                                 "required shape %r" % (ar.shape, self.shape))
        return ar


class mapping_loader(object):
    def __init__(self, repo, name_type='string', value_type=None):
        self.name_loader = repo.make_loader(name_type)
        self.value_loader = repo.make_loader(value_type)

    def __call__(self, name, dct, context):
        if not isinstance(dct, Mapping):
            raise InputDataError(
                'Expected mapping, got %r instead' % (dct))
        return dict(
            (self.name_loader(name, name, context),
                self.value_loader(name, dct[name], context))
            for name in dct)


class named_object_loader(object):
    def __init__(self, repo, members):
        self.loader = repo.make_object_loader(members)

    def __call__(self, name, input_value, context):
        k, v = list(input_value.items())[0]
        return self.loader(k, v, context)


class named_value_loader(object):
    def __init__(self, repo, value_type, name_type='string'):
        self.name_loader = repo.make_loader(name_type)
        self.value_loader = repo.make_loader(value_type)

    def __call__(self, name, input_value, context):
        name, value = list(input_value.items())[0]
        return (self.name_loader(None, name, context),
                self.value_loader(None, value, context))


_globals = globals().copy()
loaders = dict((name[:-7], _globals[name]) for name in _globals
               if name.endswith('_loader') and
               isinstance(_globals[name], type))



from .. Error import UnitsParseError


__all__ = ['units_db']


class UnitsDB(object):
    prefixes = {
        'Y': 1e24,
        'Z': 1e21,
        'E': 1e18,
        'P': 1e15,
        'T': 1e12,
        'G': 1e9,
        'M': 1e6,
        'k': 1e3,
        'h': 1e2,
        'da': 1e1,
        'd': 1e-1,
        'c': 1e-2,
        'm': 1e-3,
        'u': 1e-6,
        'n': 1e-9,
        'p': 1e-12,
        'f': 1e-15,
        'a': 1e-18,
        'z': 1e-21,
        'y': 1e-24,
    }

    def __init__(self):
        self.db = {}

    def lookup(self, name):
        # Try without prefix first
        if name in self.db:
            return self.db[name]
        # Try with single letter prefix:
        if name[1:] in self.db and name[:1] in self.prefixes:
            return self.prefixes[name[:1]]*self.db[name[1:]]
        # Try with double letter prefix (just 'da'):
        if name[2:] in self.db and name[:2] in self.prefixes:
            return self.prefixes[name[1:]]*self.db[name[2:]]
        # Raise error.
        raise UnitsParseError('Unknown units: %r' % name)

    def add(self, name, val):
        self.db[name] = val

# Define global units database.


units_db = UnitsDB()



from .. Error import RINGReaderError
from .. RDkitWrapper.MolQuery import AtomRing, AtomConnectivityAtom,\
    ConstraintNumber, MolCharge, MolAromatic, MolOlefinic, MolParaffinic,\
    MolCyclic, MolLinear, MolQuery, AtomRadical, BondQuery, BondConstraint, \
    AtomConnectivityGroup, AtomIsInRing, AtomIsAromatic, AtomIsAllylic, \
    DoubleBondStereoConstraint, AtomNRing
from rdkit.Chem import rdqueries
from rdkit import Chem
"""
Example:
from pgradd.RINGParser.Reader import Read
from rdkit import Chem
s = " \
fragment a{C labeled C1 \
C labeled C2 ring bond to C1} "
mplquery = Read(s)
mol = Chem.MolFromSmiles('C1CCCCC1')
molquery.GetQueryMatches(mol)
"""


class MolQueryReader(object):
    def __init__(self, tree, RINGgroups=None):
        # ast = Abstract Syntax Tree
        self.tree = tree
        # dict of groups where the key is equal to group name
        self.RINGgroups = RINGgroups

    def ReadBondTypeAtomConstraint(self, tree):
        return BondQuery(tree[0])

    def ReadGroupName(self, tree):
        return tree[0]

    def ReadAtomConstraintConnectivity(self, tree):
        i = 0
        # Boolean
        if tree[i][0].name == 'Boolean':
            s = tree[i][1]
            if s != '!':
                raise NotImplementedError("Unsupported Boolean operator: '"
                                          + s + "'")
            i += 1
            negate = True
        else:
            negate = False
        # ConstraintNumber
        if tree[i][0].name == 'ConstraintNumber':
            CN = ConstraintNumber(tree[i][1:])
            i += 1
        else:
            CN = ConstraintNumber('>=1')
        # Connected atom/group
        if tree[i][0].name == 'AtomType':
            t = 0
            connected, constraints = self.ReadAtomType(tree[i][1:])
            # rdkit atom object
            i += 1
        elif tree[i][0].name == 'GroupName':
            t = 1
            try:
                connected = self.RINGgroups[self.ReadGroupName(tree[i][1:])]
            except KeyError:
                raise RINGReaderError("Unrecognized group name :'"
                                      + self.ReadGroupName(tree[i][1:])+"'")
            i += 1
            # string
        # Bond type
        if len(tree) == i+1:
            assert tree[i][0].name == 'BondType'
            bondquery = self.ReadBondTypeAtomConstraint(tree[i][1:])
        else:
            bondquery = BondQuery('single')
        if t == 0:
            return AtomConnectivityAtom(negate, CN, connected,
                                        bondquery, constraints)
        elif t == 1:
            return AtomConnectivityGroup(negate, CN, connected,
                                         bondquery)

    def ReadAtomConstraintRing(self, tree):
        i = 0
        # Boolean
        if tree[i][0].name == 'Boolean':
            s = tree[i][1]
            if s != '!':
                raise NotImplementedError("Unsupported Boolean operator: '"
                                          + s + "'")
            i += 1
            negate = True
        else:
            negate = False
        assert tree[i][0].name == 'ConstraintNumber'
        CN = ConstraintNumber(tree[i][1:])
        return AtomRing(negate, CN)

    def ReadAtomConstraintRadical(self, tree):
        i = 0
        # Boolean
        if tree[i][0].name == 'Boolean':
            s = tree[i][1]
            if s != '!':
                raise NotImplementedError("Unsupported Boolean operator: '"
                                          + s + "'")
            i += 1
            negate = True
        else:
            negate = False
        assert tree[i][0].name == 'ConstraintNumber'
        CN = ConstraintNumber(tree[i][1:])
        return AtomRadical(negate, CN)

    def ReadAtomConstraintNRing(self, tree):
        i = 0
        # Boolean
        if tree[i][0].name == 'Boolean':
            s = tree[i][1]
            if s != '!':
                raise NotImplementedError("Unsupported Boolean operator: '"
                                          + s + "'")
            i += 1
            negate = True
        else:
            negate = False
        assert tree[i][0].name == 'ConstraintNumber'
        CN = ConstraintNumber(tree[i][1:])
        return AtomNRing(negate, CN)

    def ReadAtomConstraints(self, tree):
        assert tree[0][0].name in ('AtomConstraintConnectivity',
                                   'AtomConstraintRing',
                                   'AtomConstraintRadical',
                                   'AtomConstraintNRing')
        if tree[0][0].name == 'AtomConstraintConnectivity':
            return self.ReadAtomConstraintConnectivity(tree[0][1:])
        elif tree[0][0].name == 'AtomConstraintRing':
            return self.ReadAtomConstraintRing(tree[0][1:])
        elif tree[0][0].name == 'AtomConstraintRadical':
            return self.ReadAtomConstraintRadical(tree[0][1:])
        elif tree[0][0].name == 'AtomConstraintNRing':
            return self.ReadAtomConstraintNRing(tree[0][1:])

    def ReadAtomConstraintChain(self, tree, molquery, idx):
        assert tree[0][0].name == 'AtomConstraints'
        constraint = self.ReadAtomConstraints(tree[0][1:])
        molquery.AppendAtomConstraint(constraint, idx)

        if len(tree) > 1:
            assert tree[1][0].name == 'AtomConstraintChain'
            self.ReadAtomConstraintChain(tree[1][1:], molquery, idx)

    def ReadAtomSuffix(self, tree, atom):
        constraint = None
        # '+','-','.',':','+.','-.','*'
        if tree[0] == '+.':
            constraint = AtomRadical(False, ConstraintNumber('=1'))
            atom.ExpandQuery(rdqueries.FormalChargeEqualsQueryAtom(1))
        elif tree[0] == '-.':
            constraint = AtomRadical(False, ConstraintNumber('=1'))
            atom.ExpandQuery(rdqueries.FormalChargeEqualsQueryAtom(-1))
        elif tree[0] == '+':
            atom.ExpandQuery(rdqueries.FormalChargeEqualsQueryAtom(1))
        elif tree[0] == '-':
            atom.ExpandQuery(rdqueries.FormalChargeEqualsQueryAtom(-1))
        elif tree[0] == '.':
            constraint = AtomRadical(False, ConstraintNumber('=1'))
        elif tree[0] == ':':
            constraint = AtomRadical(False, ConstraintNumber('=2'))
        elif tree[0] == ':.':
            constraint = AtomRadical(False, ConstraintNumber('=3'))
        elif tree[0] == '*':
            from rdkit.Chem import GetPeriodicTable
            atomicnum = atom.GetAtomicNum()
            atom = rdqueries.AtomNumEqualsQueryAtom(atomicnum)
            valence = GetPeriodicTable().GetDefaultValence(atomicnum)
            atom.ExpandQuery(rdqueries.TotalValenceEqualsQueryAtom(valence+1))
            atom.ExpandQuery(rdqueries.FormalChargeEqualsQueryAtom(1))
        elif tree[0] == '?':
            pass
        else:
            s = "Unsupported atom suffic: '" + tree[0] + "'"
            raise NotImplementedError(s)
        return constraint

    def ReadSymbols(self, tree):
        if tree[0] in ['any atom', '$']:
            atom = rdqueries.AtomNumGreaterQueryAtom(0)
        elif tree[0] in ['heteroatom', '&']:
            # N, O, P, S
            atom = rdqueries.AtomNumEqualsQueryAtom(7)
            atom.ExpandQuery(rdqueries.AtomNumEqualsQueryAtom(8),
                             how=Chem.rdchem.CompositeQueryType.COMPOSITE_OR)
            atom.ExpandQuery(rdqueries.AtomNumEqualsQueryAtom(15),
                             how=Chem.rdchem.CompositeQueryType.COMPOSITE_OR)
            atom.ExpandQuery(rdqueries.AtomNumEqualsQueryAtom(16),
                             how=Chem.rdchem.CompositeQueryType.COMPOSITE_OR)
        elif tree[0] in ['heavy atom', 'X']:
            # heavier than H
            atom = rdqueries.AtomNumGreaterQueryAtom(1)
        elif tree[0][0].islower():
            # aromatic molecule
            symbol = tree[0][0].upper()+tree[0][1:]
            try:
                atom = Chem.Atom(symbol)
                atom.SetIsAromatic(True)
            except RuntimeError:
                msg = 'Element aromatic ' + symbol + ' not found'
                raise RINGReaderError(msg)
        elif tree[0] == 'M':
            # metal
            atom = rdqueries.AtomNumGreaterQueryAtom(19)
        else:
            try:
                atom = Chem.Atom(tree[0])
                atom = rdqueries.AtomNumEqualsQueryAtom(atom.GetAtomicNum())
            except RuntimeError:
                msg = 'Element ' + tree[0] + ' not found'
                raise RINGReaderError(msg)
        return atom

    def ReadAtomPrefix(self, tree):
        if tree[0] == 'aromatic':
            return AtomIsAromatic(negate=False)
        elif tree[0] == 'nonaromatic':
            return AtomIsAromatic(negate=True)
        elif tree[0] == 'ringatom':
            return AtomIsInRing(negate=False)
        elif tree[0] == 'nonringatom':
            return AtomIsInRing(negate=True)
        elif tree[0] == 'allylic':
            return AtomIsAllylic(negate=False)

    def ReadAtomType(self, tree):
        i = 0
        constraints = list()
        if tree[i][0].name == 'AtomPrefix':
            constraints.append(self.ReadAtomPrefix(tree[i][1:]))
            i += 1

        assert tree[i][0].name == 'Symbols'
        atom = self.ReadSymbols(tree[i][1:])
        i += 1

        if len(tree) > i:
            constraint = self.ReadAtomSuffix(tree[i][1:], atom)
            if constraint:
                constraints.append(constraint)
        else:
            # no suffix means it's neutral. RDkit atom with 0 formal charge and
            # radical electron finds all variables of formal charge and radical
            # electrons, so they need to be treated.
            atom.ExpandQuery(rdqueries.FormalChargeEqualsQueryAtom(0))
            constraints.append(AtomRadical(False, ConstraintNumber('=0')))
        # return atom
        return atom, constraints

    def ReadAtom(self, tree, molquery):
        assert tree[0][0].name == 'AtomType'
        atom, constraints = self.ReadAtomType(tree[0][1:])
        idx = molquery.mol.AddAtom(atom)
        if constraints:
            for constraint in constraints:
                molquery.AppendAtomConstraint(constraint, idx)
        assert tree[1][0].name == 'AtomLabel'
        molquery.atom_names.append(tree[1][1])
        if len(tree) > 2:
            assert tree[2][0].name == 'AtomConstraintChain'
            self.ReadAtomConstraintChain(tree[2][1:], molquery, idx)

    def ReadBondTypeBondedAtom(self, idx, idx_connected, bondtype, molquery):
        if bondtype == 'single':
            molquery.mol.AddBond(idx, idx_connected, Chem.BondType.SINGLE)
        elif bondtype == 'double':
            molquery.mol.AddBond(idx, idx_connected, Chem.BondType.DOUBLE)
        elif bondtype == 'triple':
            molquery.mol.AddBond(idx, idx_connected, Chem.BondType.TRIPLE)
        elif bondtype == 'quadruple':
            molquery.mol.AddBond(idx, idx_connected, Chem.BondType.QUADRUPLE)
        elif bondtype == 'ring':
            molquery.mol.AddBond(idx, idx_connected, Chem.BondType.UNSPECIFIED)
            molquery.AppendBondConstraint(idx, idx_connected,
                                          BondConstraint(BondQuery(bondtype)))
        elif bondtype == 'nonring':
            molquery.mol.AddBond(idx, idx_connected, Chem.BondType.UNSPECIFIED)
            molquery.AppendBondConstraint(idx, idx_connected,
                                          BondConstraint(BondQuery(bondtype)))
        elif bondtype == 'aromatic':
            molquery.mol.AddBond(idx, idx_connected, Chem.BondType.AROMATIC)
        elif bondtype == 'any':
            molquery.mol.AddBond(idx, idx_connected, Chem.BondType.UNSPECIFIED)
            return True
        elif bondtype == 'strong':
            molquery.mol.AddBond(idx, idx_connected, Chem.BondType.UNSPECIFIED)
            molquery.AppendBondConstraint(idx, idx_connected,
                                          BondConstraint(BondQuery(bondtype)))
        elif bondtype == 'partial':
            molquery.mol.AddBond(idx, idx_connected, Chem.BondType.UNSPECIFIED)
            molquery.AppendBondConstraint(idx, idx_connected,
                                          BondConstraint(BondQuery(bondtype)))
        else:
            raise NotImplementedError("Unsupported bond type: '"+bondtype+"'")

    def ReadBondedAtom(self, tree, molquery):
        assert tree[0][0].name == 'AtomType'
        atom, constraints = self.ReadAtomType(tree[0][1:])
        idx = molquery.mol.AddAtom(atom)
        if constraints:
            for constraint in constraints:
                molquery.AppendAtomConstraint(constraint, idx)

        assert tree[1][0].name == 'AtomLabel'
        if tree[1][0].name in molquery.atom_names:
            raise RINGReaderError('Atom Label ' + tree[1][0]
                                  + ' is alreadyd declared!')
        molquery.atom_names.append(tree[1][1])
        assert tree[2][0].name == 'BondType'
        bondtype = tree[2][1:][0]
        assert tree[3][0].name == 'AtomLabel'
        try:
            idx_connected = molquery.atom_names.index(tree[3][1])
        except Exception:
            msg = 'Atom Label '+tree[3][1]+' not found'
            raise RINGReaderError(msg)
        self.ReadBondTypeBondedAtom(idx, idx_connected,
                                    bondtype, molquery)

        if len(tree) > 4:
            assert tree[4][0].name == 'AtomConstraintChain'
            self.ReadAtomConstraintChain(tree[4][1:], molquery, idx)

    def ReadRingBond(self, tree, molquery):
        assert tree[0][0].name == 'AtomLabel'
        try:
            idx1 = molquery.atom_names.index(tree[0][1])
        except Exception:
            msg = 'Atom Label '+tree[0][1]+' not found'
            raise RINGReaderError(msg)

        assert tree[1][0].name == 'BondType'
        bondtype = tree[1][1:][0]
        assert tree[2][0].name == 'AtomLabel'
        try:
            idx2 = molquery.atom_names.index(tree[2][1])
        except Exception:
            msg = 'Atom Label '+tree[2][1]+' not found'
            raise RINGReaderError(msg)
        self.ReadBondTypeBondedAtom(idx1, idx2, bondtype, molquery)

    def ReadStereoDoubleBond(self, tree, molquery):
        i = 0
        assert tree[i][0].name == 'AtomLabel'
        try:
            idx1 = molquery.atom_names.index(tree[i][1])
            i += 1
        except Exception:
            msg = 'Atom Label '+tree[i][1]+' not found'
            raise RINGReaderError(msg)

        # Boolean
        if tree[i][0].name == 'Boolean':
            s = tree[i][1]
            if s != '!':
                raise NotImplementedError("Unsupported Boolean operator: '"
                                          + s + "'")
            i += 1
            negate = True
        else:
            negate = False

        assert tree[i][0].name == 'DoubleBondStereoType'
        if tree[i][1:][0] == 'cis':
            stereobondtype = Chem.rdchem.BondStereo.STEREOZ
        elif tree[i][1:][0] == 'trans':
            stereobondtype = Chem.rdchem.BondStereo.STEREOE
        elif tree[i][1:][0] == 'notspecified':
            stereobondtype = Chem.rdchem.BondStereo.STEREONONE
        else:
            raise NotImplementedError("Unsupported stereo type: '"
                                      + tree[i][1:][0]+"'")
        i += 1
        assert tree[i][0].name == 'AtomLabel'
        try:
            idx2 = molquery.atom_names.index(tree[i][1])
            i += 1
        except Exception:
            msg = 'Atom Label ' + tree[i][1] + ' not found'
            raise RINGReaderError(msg)
        assert tree[i][0].name == 'AtomLabel'
        try:
            idx3 = molquery.atom_names.index(tree[i][1])
            i += 1
        except Exception:
            msg = 'Atom Label '+tree[i][1]+' not found'
            raise RINGReaderError(msg)
        assert tree[i][0].name == 'AtomLabel'
        try:
            idx4 = molquery.atom_names.index(tree[i][1])
            i += 1
        except Exception:
            msg = 'Atom Label '+tree[i][1]+' not found'
            raise RINGReaderError(msg)
        # Error checks double bond check
        dbond = molquery.mol.GetBondBetweenAtoms(idx3, idx4)
        if not dbond:
            msg = 'No bond found between '+tree[i-2][1]+' and '+tree[i-1][1]
            raise RINGReaderError(msg)
        elif dbond.GetBondType() != Chem.BondType.DOUBLE:
            msg = 'Bond between ' + tree[i-2][1] + ' and '\
                + tree[i-1][1] + ' is not double bond'
            raise RINGReaderError(msg)
        # Check connectivity
        bond13 = molquery.mol.GetBondBetweenAtoms(idx1, idx3)
        bond14 = molquery.mol.GetBondBetweenAtoms(idx1, idx4)
        bond23 = molquery.mol.GetBondBetweenAtoms(idx2, idx3)
        bond24 = molquery.mol.GetBondBetweenAtoms(idx2, idx4)
        if bond13 and bond23 or bond14 and bond24:
            msg = tree[0][1] + ' and ' + tree[i-3][1] +\
                ' are bond to the same atom (double bond stereo)'
            raise RINGReaderError(msg)
        elif not bond13 and not bond14:
            msg = tree[0][1] + ' is not bound to ' + tree[i-2][1] +\
                ' or ' + tree[i-1][1] + '(double bond stereo)'
            raise RINGReaderError(msg)
        elif not bond23 and not bond24:
            msg = tree[i-3][1] + ' is not bound to ' + tree[i-2][1] +\
                ' or ' + tree[i-1][1] + '(double bond stereo)'
            raise RINGReaderError(msg)
        molquery.AppendDoubleBondStereoConstraint(idx1, idx2, idx3, idx4,
                                                  DoubleBondStereoConstraint
                                                  (negate, stereobondtype))

    def ReadAtomChain(self, tree, molquery):
        assert tree[0][0].name in ['BondedAtom',
                                   'RingBond',
                                   'StereoDoubleBond']
        if tree[0][0].name == 'BondedAtom':
            self.ReadBondedAtom(tree[0][1:], molquery)
        elif tree[0][0].name == 'RingBond':
            self.ReadRingBond(tree[0][1:], molquery)
        elif tree[0][0].name == 'StereoDoubleBond':
            self.ReadStereoDoubleBond(tree[0][1:], molquery)

        if len(tree) > 1:
            assert tree[1][0].name == 'AtomChain'
            self.ReadAtomChain(tree[1][1:], molquery)

    def ReadMolQuery(self, tree, molquery):
        assert tree[0][0].name == 'Atom'
        self.ReadAtom(tree[0][1:], molquery)

        if len(tree) > 1:
            assert tree[1][0].name == 'AtomChain'
            self.ReadAtomChain(tree[1][1:], molquery)

    def ReadMolQueryPrefix(self, tree, molquery):
        i = 0
        if tree[i] == 'positive':
            molquery.AppendMolConstraint(MolCharge(ConstraintNumber('=1')))
            i += 1
        elif tree[i] == 'negative':
            molquery.AppendMolConstraint(MolCharge(ConstraintNumber('=-1')))
            i += 1
        elif tree[i] == 'neutral':
            molquery.AppendMolConstraint(MolCharge(ConstraintNumber('=0')))
            i += 1

        if i < len(tree):
            if tree[i] == 'aromatic':
                molquery.AppendMolConstraint(MolAromatic())
                i += 1
            elif tree[i] == 'olefinic':
                molquery.AppendMolConstraint(MolOlefinic())
                i += 1
            elif tree[i] == 'paraffinic':
                molquery.AppendMolConstraint(MolParaffinic())
                i += 1

        if i < len(tree):
            if tree[i] == 'cyclic':
                molquery.AppendMolConstraint(MolCyclic())
            elif tree[i] == 'linear':
                molquery.AppendMolConstraint(MolLinear())
            else:
                raise NotImplementedError('Unsupported mol prefix: %s'
                                          % tree[0])

    def Read(self):
        # initialize Enhanced MolQuery
        molquery = MolQuery()
        # prefix
        assert self.tree[0][0].name == 'Prefix'
        if len(self.tree[0]) != 1:
            self.ReadMolQueryPrefix(self.tree[0][1:], molquery)
        # fragment name
        assert self.tree[1][0].name in ['FragmentName',
                                        'ReactantName',
                                        'GroupName']
        molquery.name = self.tree[1][1]
        # start building molecule
        assert self.tree[2][0].name == 'MolQuery'
        self.ReadMolQuery(self.tree[2][1:], molquery)
        # return
        if not molquery.mol:
            raise RINGReaderError('Rdkit mol is not created')
        return molquery



from ..Error import ReactionQueryError
from rdkit import Chem
from . MolQuery import MolQuery
from collections import defaultdict
from itertools import product as itpd
from rdkit.Chem import rdqueries
from operator import add
"""
Written by Geun Ho Gu 8/4/2016
ReactionQuery class is equipped with enhanced query comparison over
RDkit query comparison
This module uses a graph description language called R.I.N.G.
Rangarajan, S., Kaminski, T., Van Wyk, E., Bhan A., and Daoutidis, P.,
"Language-oriented rule-based reaction network generation and analysis:
    Algorithms of RING", Computers and Chemical Engineering 64 (2014)
    124, 10.1016/j.compchemeng.2014.02.007
Rangarajan, S., Bhan, A., and Daoutidis, P.,
"Language-oriented rule-based reaction network generation and analysis:
    Descrpition of RING", Computers and Chemical Engineering 45 (2012) 114,
    10.1016/j.compchemeng.2012.06.008
"""


class transformation(object):
    pass


class BondForm(transformation):
    def __init__(self, idx1, idx2, bondtype):
        self.idx1 = idx1
        self.idx2 = idx2
        self.bondtype = bondtype

    def __call__(self, comb_mol, mapped_index):
        comb_mol.AddBond(mapped_index[self.idx1],
                         mapped_index[self.idx2],
                         self.bondtype)


class BondBreak(transformation):
    def __init__(self, idx1, idx2):
        self.idx1 = idx1
        self.idx2 = idx2

    def __call__(self, comb_mol, mapped_index):
        comb_mol.RemoveBond(mapped_index[self.idx1], mapped_index[self.idx2])


class BondModify(transformation):
    def __init__(self, idx1, idx2, bondtype):
        self.idx1 = idx1
        self.idx2 = idx2
        self.bondtype = bondtype

    def __call__(self, comb_mol, mapped_index):
        comb_mol.RemoveBond(mapped_index[self.idx1], mapped_index[self.idx2])
        comb_mol.AddBond(mapped_index[self.idx1],
                         mapped_index[self.idx2], self.bondtype)


class BondIncrease(transformation):
    def __init__(self, idx1, idx2):
        self.idx1 = idx1
        self.idx2 = idx2

    def getbondtype(self, bondtype):
        if bondtype.GetBondType().__str__() == 'SINGLE':
            return Chem.BondType().DOUBLE
        elif bondtype.GetBondType().__str__() == 'DOUBLE':
            return Chem.BondType().TRIPLE
        elif bondtype.GetBondType().__str__() == 'TRIPLE':
            return Chem.BondType().QUADRUPLE
        elif bondtype.GetBondType().__str__() == 'QUADRUPLE':
            return Chem.BondType().QUINTUPLE
        elif bondtype.GetBondType().__str__() == 'QUINTUPLE':
            raise ReactionQueryError('BondIncrease: Increasing bond order',
                                     'above quintuple bond is not supported')
        elif bondtype.GetBondType().__str__() == 'AROMATIC':
            raise ReactionQueryError('BondIncrease: Increasing bond order',
                                     'of aromatic bond is not supported')
        elif bondtype.GetBondType().__str__() == 'DATIVE':
            raise ReactionQueryError('BondIncrease: Increasing bond order of',
                                     'partial bond is not supported')
        else:
            raise ReactionQueryError("BondDecrease: Unsupported bond type:'"
                                     + bondtype.GetBondType().__str__() + "'")

    def __call__(self, comb_mol, mapped_index):
        bondtype = self.getbondtype(comb_mol.
                                    GetBondBetweenAtoms(mapped_index
                                                        [self.idx1],
                                                        mapped_index
                                                        [self.idx2]))
        comb_mol.RemoveBond(mapped_index[self.idx1], mapped_index[self.idx2])
        comb_mol.AddBond(mapped_index[self.idx1],
                         mapped_index[self.idx2],
                         bondtype)


class AtomTypeModify(transformation):
    def __init__(self, idx, radical, charge, valence):
        self.idx = idx
        self.radical = radical
        self.charge = charge
        self.valence = valence

    def __call__(self, comb_mol, mapped_index):
        atom = comb_mol.GetAtomWithIdx(mapped_index[self.idx])
        if self.valence != 0:
            from rdkit.Chem import GetPeriodicTable
            atomicnum = atom.GetAtomicNum()
            atom = rdqueries.AtomNumEqualsQueryAtom(atomicnum)
            valence = GetPeriodicTable().GetDefaultValence(atomicnum)
            atom.ExpandQuery(rdqueries.
                             TotalValenceEqualsQueryAtom(valence+self.valence))
            atom.ExpandQuery(rdqueries.
                             FormalChargeEqualsQueryAtom(self.charge))
            comb_mol.ReplaceAtom(self.idx, atom)

        atom.SetNumRadicalElectrons(self.radical)
        atom.SetFormalCharge(self.charge)


class RadicalIncrease(transformation):
    def __init__(self, idx):
        self.idx = idx

    def __call__(self, comb_mol, mapped_index):
        atom = comb_mol.GetAtomWithIdx(mapped_index[self.idx])
        atom.SetNumRadicalElectrons(atom.GetNumRadicalElectrons()+1)


class RadicalDecrease(transformation):
    def __init__(self, idx):
        self.idx = idx

    def __call__(self, comb_mol, mapped_index):
        atom = comb_mol.GetAtomWithIdx(mapped_index[self.idx])
        atom.SetNumRadicalElectrons(atom.GetNumRadicalElectrons()-1)


class ChargeIncrease(transformation):
    def __init__(self, idx):
        self.idx = idx

    def __call__(self, comb_mol, mapped_index):
        atom = comb_mol.GetAtomWithIdx(mapped_index[self.idx])
        atom.SetFormalCharge(atom.GetFormalCharge()+1)


class ChargeDecrease(transformation):
    def __init__(self, idx):
        self.idx = idx

    def __call__(self, comb_mol, mapped_index):
        atom = comb_mol.GetAtomWithIdx(mapped_index[self.idx])
        atom.SetFormalCharge(atom.GetFormalCharge()-1)


class BondDecrease(transformation):
    def __init__(self, idx1, idx2):
        self.idx1 = idx1
        self.idx2 = idx2

    def getbondtype(self, bondtype):
        if bondtype.GetBondType().__str__() == 'SINGLE':
            return None
        elif bondtype.GetBondType().__str__() == 'DOUBLE':
            return Chem.BondType().SINGLE
        elif bondtype.GetBondType().__str__() == 'TRIPLE':
            return Chem.BondType().DOUBLE
        elif bondtype.GetBondType().__str__() == 'QUADRUPLE':
            return Chem.BondType().TRIPLE
        elif bondtype.GetBondType().__str__() == 'QUINTUPLE':
            return Chem.BondType().QUADRUPLE
        elif bondtype.GetBondType().__str__() == 'AROMATIC':
            raise ReactionQueryError('BondDecrease: Decreasing bond order',
                                     'of aromatic bond is not supported')
        elif bondtype.GetBondType().__str__() == 'DATIVE':
            raise ReactionQueryError('BondDecrease: Decreasing bond order',
                                     'of partial bond is not supported')
        else:
            raise ReactionQueryError("BondDecrease: Unsupported bond type:'"
                                     + bondtype.GetBondType().__str__() + "'")

    def __call__(self, comb_mol, mapped_index):
        bondtype = self.getbondtype(comb_mol.
                                    GetBondBetweenAtoms(mapped_index
                                                        [self.idx1],
                                                        mapped_index
                                                        [self.idx2]))
        comb_mol.RemoveBond(mapped_index[self.idx1],
                            mapped_index[self.idx2])
        if bondtype:
            comb_mol.AddBond(mapped_index[self.idx1],
                             mapped_index[self.idx2],
                             bondtype)


class ReactionQuery(object):
    def __init__(self):
        self.mol = Chem.RWMol(Chem.Mol())
        self.transformations = list()
        self.reactantquery = dict()
        self.constraints = defaultdict(list)
        self.atom_names = list()

    def __repr__(self):
        s = "%s(" % (self.__class__.__name__)
        if hasattr(self, 'name'):
            s += "name='%s', " % (self.name)
        s += "NumberReactants='%s, " % (len(self.reactants))
        nconstraint = len(self.mol_constraints)
        for constraint in self.atom_constraints:
            nconstraint += len(self.atom_constraints[constraint])
        s += "NumberConstraints=('%s')" % (nconstraint)
        s += ")"
        return s

    def AppendReactantQuery(self, reactant):
        if not isinstance(reactant, (Chem.Mol, MolQuery)):
            raise ReactionQueryError("Unrecognized instance'"
                                     + type(reactant)+"'")
        self.reactantquery[reactant.name] = reactant
        self.atom_names += reactant.atom_names

    def GetNumReactantTemplates(self):
        return len(self.reactantquery)

    def RunReactants(self, reactants):
        if isinstance(reactants, Chem.Mol):
            reactants = [reactants]
        elif isinstance(reactants, tuple):
            reactants = list(reactants)
        # Error test
        if len(reactants) != len(self.reactantquery):
            raise ReactionQueryError('Number of reactant mis-match')
        for i in range(0, len(reactants)):
            if not isinstance(reactants[i], Chem.Mol):
                raise ReactionQueryError("Unrecognized instance'"
                                         + type(reactants[i])+"'")
            reactants[i] = Chem.AddHs(reactants[i])
        # make matrix of matching index
        self.match_indexes = list()
        i = 0
        for reactant_name in self.reactantquery:
            if isinstance(self.reactantquery[reactant_name], Chem.Mol):
                match = reactants[i].GetSubstructMatches(self.reactantquery
                                                         [reactant_name])
                if not match:
                    return tuple()
                else:
                    self.match_indexes.append(match)
            elif isinstance(self.reactantquery[reactant_name], MolQuery):
                match = self.reactantquery[reactant_name].\
                    GetQueryMatches(reactants[i])
                if not match:
                    return tuple()
                else:
                    self.match_indexes.append(match)
            i += 1
        # make a index for combined reactants mol object
        self.combined_mol_match_index = list()
        for match in itpd(*self.match_indexes):
            row = list()
            modifier = 0
            for i in range(0, len(match)):
                if i != 0:
                    modifier = reactants[i-1].GetNumAtoms()
                row += list(map(add, list(match[i]), [modifier]*len(match[i])))
            self.combined_mol_match_index.append(row)
        # Combine reactants
        self.combined_mol = reactants[0]
        for i in range(1, len(reactants)):
            self.combined_mol = Chem.CombineMols(self.combined_mol,
                                                 reactants[i])

        self.combined_mol = Chem.RWMol(self.combined_mol)
        # Transform!!
        product_list = list()
        for matches in self.combined_mol_match_index:
            products = self.combined_mol.__copy__()
            for transform in self.transformations:
                transform(products, matches)
            products = Chem.GetMolFrags(products,
                                        asMols=True,
                                        sanitizeFrags=False)
            product_list.append(products)
        return tuple(product_list)

    def ModifyMolInPlace(self, reactants):
        # This applies transformation for all the matching pattern found in mol
        if len(self.reactantquery) != 1:
            raise ReactionQueryError('ModifyMolInPlace: Only usable for',
                                     'monomolecular transformation')
        if isinstance(reactants, Chem.Mol):
            reactants = [reactants]
        elif isinstance(reactants, tuple):
            reactants = list(reactants)
        # Error test
        if len(reactants) != len(self.reactantquery):
            raise ReactionQueryError('Number of reactant mis-match')
        for i in range(0, len(reactants)):
            if not isinstance(reactants[i], Chem.Mol):
                raise ReactionQueryError("Unrecognized instance'"
                                         + type(reactants[i])+"'")
            reactants[i] = Chem.AddHs(reactants[i])
        # make matrix of matching index
        self.match_indexes = list()
        i = 0
        for reactant_name in self.reactantquery:
            if isinstance(self.reactantquery[reactant_name], Chem.Mol):
                match = reactants[i].GetSubstructMatches(self.reactantquery
                                                         [reactant_name])
                if not match:
                    return tuple()
                else:
                    self.match_indexes.append(match)
            elif isinstance(self.reactantquery[reactant_name], MolQuery):
                match = self.reactantquery[reactant_name].\
                    GetQueryMatches(reactants[i])
                if not match:
                    return tuple()
                else:
                    self.match_indexes.append(match)
            i += 1
        # make a index for combined reactants mol object
        self.combined_mol_match_index = list()
        for match in itpd(*self.match_indexes):
            row = list()
            modifier = 0
            for i in range(0, len(match)):
                if i != 0:
                    modifier = reactants[i-1].GetNumAtoms()
                row += list(map(add,
                                list(match[i]),
                                [modifier]*len(match[i])))
            self.combined_mol_match_index.append(row)
        # Combine reactants
        self.combined_mol = reactants[0]
        for i in range(1, len(reactants)):
            self.combined_mol = Chem.CombineMols(self.combined_mol,
                                                 reactants[i])

        self.combined_mol = Chem.RWMol(self.combined_mol)
        # Transform!!        product_list = list()
        for matches in self.combined_mol_match_index:
            for transform in self.transformations:
                transform(self.combined_mol, matches)
        return self.combined_mol



import os
import sys

# What environment variable is used to override use of the bundled
# data directory?
_data_dir_envvar = 'pgradd_DATA_DIR'

# We'll cache the located data directory root in this variable:
_data_dir_cached = False


def get_data_dir():
    """Returns the directory containing the pgradd data library.

    Return the cached directory path if it has been set.  Otherwise:

    1. If the _data_dir_envvar is set in the runtime environment,
       cache and use it as the data dir

    2. This Python file exists somewhere under the 'pgradd' module
       directory; step back until we get a path that ends with '/pgradd'
       and tack '/data' onto the end of it.  Voila, our bundled data
       directory.

    The caveat to 2. is when this script is run from the GroupAdd
    directory for testing purposes; the __file__ will be just
    "DataDir.py" so it needs to be made absolute.
    """
    global _data_dir_cached, _data_dir_cached

    if not _data_dir_cached:
        base_path = os.getenv(_data_dir_envvar, False)
        if not base_path:
            base_path = os.path.abspath(os.path.dirname(__file__))
            while base_path != os.sep and not base_path.endswith(os.sep +
                                                                 'pgradd'):
                base_path = os.path.dirname(base_path)
            if base_path == os.sep:
                raise RuntimeError('DataDir.get_data_dir: unable to locate',
                                   'bundled data directory')
            base_path = os.path.join(base_path, 'data')
        if not os.path.isdir(base_path):
            raise RuntimeError('DataDir.get_data_dir: directory does not',
                               'exist: ' + base_path)
        _data_dir_cached = base_path
    return _data_dir_cached

#
# This code can be tested very simply by
#
#   python DataDir.py
#


if __name__ == '__main__':
    sys.stdout.write('pgradd data directory = {0:s}'.format(get_data_dir()))



# 2016. Vlachos Group Geun Ho Gu. University of Delaware.
"""
==========
Exceptions
==========

All the exceptions are listed here.

"""
import re
drawmolonerror = 0
if drawmolonerror:
    from .DrawMol import moltosvg

__all__ = []


class SemiEmpiricalMethodError(Exception):
    """ errors related to semi-empirical methods """


class GroupSyntaxError(Exception):
    """Exception raised when the group parser encounters invalid input."""


class GroupSchemeError(Exception):
    """Exception raised when a group scheme-related error occurs."""


class GroupMissingDataError(Exception):
    """
    Exception raised when a group library lacks the requested data for
    some group.

    Attributes
    ----------
    groups : list of :class:`chemtk.groupadd.Group`
        Groups for which data are missing.
    property_set_name : str
        Name of property set for which data are missing.
    """
    def __init__(self, groups, property_set_name):
        self.groups = groups
        self.property_set_name = property_set_name

    def __str__(self):
        if len(self.groups) == 1:
            return ('Library does not define property set %r for group %s'
                    % (str(self.property_set_name), str(self.groups[0])))
        else:
            return ('Library does not define property set %r for groups { %s }'
                    % (str(self.property_set_name),
                       ', '.join(repr(str(group)) for group in self.groups)))


class PatternMatchError(Exception):
    """
    Exception raised when no known pattern matches part of a chemical
    structure.

    The atom and (possibly) the bond at which the matching failed are stored
    as attributes :attr:`atom` and :attr:`bond`.

    For center pattern matches, :attr:`atom` specifies the atom at which the
    matching failed and :attr:`bond` is None.

    For peripheral pattern matches, :attr:`atom` is the atom adjacent to
    the bond (:attr:`bond`), that is being considered in the peripheral match.

    Attributes
    ----------
    atom : :class:`chemtk.structure.Atom`
        The afflicted atom
    bond : :class:`chemtk.structure.Bond`
        The afflicted bond (possibly None)
    """
    def __init__(self, mesg, atom):
        self.mesg = mesg
        self.atom = atom
        self.mol = atom.GetOwningMol()
        self.visualize()

    def visualize(self, *args, **kwargs):
        if drawmolonerror:
            moltosvg(self.mol, highlight=[self.atom.GetIdx()], kekulize=False)

    def __str__(self):
        return '%s at atom number %s' % (self.mesg, self.atom.GetIdx())


__all__ += ['SemiEmpiricalMethodError', 'GroupSyntaxError', 'GroupSchemeError',
            'GroupMissingDataError', 'PatternMatchError']


# RING-related errors.
class RINGError(Exception):
    """Base exception for Parser errors.

    All other Parser-related exceptions inherit from this.
    """


class RINGSyntaxError(RINGError):
    """
    Exception raised due to invalid (but syntactically correct) input.
    """
    _parse = re.compile('[\n]')

    def __init__(self, tok, lineno, colno, stream):
        self.toks = set([tok])
        self.lineno = lineno
        self.colno = colno
        self.stream = stream

    def update(self, other):
        if (self.lineno < other.lineno or
                (self.lineno == other.lineno and self.colno < other.colno)):
            self.lineno = other.lineno
            self.colno = other.colno
            self.toks = other.toks.copy()
        elif (self.lineno == other.lineno and self.colno == other.colno):
            self.toks |= other.toks

    def __str__(self):
        s = 'Expected ' + ' | '.join(
            sorted(str(tok) for tok in self.toks if tok))\
            + ' at line %d column %d:\n' % (self.lineno, self.colno)
        s += self._parse.split(self.stream)[self.lineno-1] + '\n'
        s += ' '*(self.colno-1) + '^\n'
        return s

    def __repr__(self):
        return '%s(%r, %r, %r, %r)' % (
            type(self).__name__, self.toks, self.lineno, self.colno,
            self.smiles)


class RINGReaderError(RINGError):
    """
    Exception raised when input does not conform to RING syntax.
    """
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message

    def __repr__(self):
        return '%s(%r, %r)' % (type(self).__name__, self.message)


class MolQueryError(Exception):
    """
    Exception raised when input does not conform to RING syntax.
    """
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message

    def __repr__(self):
        return '%s(%r, %r)' % (type(self).__name__, self.message)


class ReactionQueryError(Exception):
    """
    Exception raised when input does not conform to RING syntax.
    """
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message

    def __repr__(self):
        return '%s(%r, %r)' % (type(self).__name__, self.message)


__all__ += ['RINGError', 'RINGSyntaxError', 'RINGReaderError',
            'MolQueryError', 'ReactionQueryError']


# Units errors.
class UnitsParseError(Exception):
    """Error from providing bad input to the units parser."""


class UnitsError(Exception):
    """Error from operation involving incompatible physical units."""


__all__ += ['UnitsParseError', 'UnitsError']


# Generic errors.
class OutsideCorrelationError(Exception):
    """Error from attempt to evaluate correlation outside its valid range."""


class ReadOnlyDataError(Exception):
    """Error raised by attempt to modify read-only data."""


class IncompleteDataError(Exception):
    """Error raised when a computation requires more data than is available."""


class IncompleteDataWarning(Warning):
    """
    Warning issued when a computation proceeds using less data than is optimal.

    This is raised when the absence of certain non-required data may lead to
    pontetially severe assumptions in later computations.
    """


__all__ += ['OutsideCorrelationError', 'ReadOnlyDataError',
            'IncompleteDataError', 'IncompleteDataWarning']



from collections import defaultdict
from rdkit import Chem
from .. RINGParser import Read
import yaml
import os
from .. Error import PatternMatchError
from . Group import Group
from . DataDir import get_data_dir

# Scheme contains the group pattern information, and can assign groups
# based on the smiles.


class Scheme(object):
    pass


class GroupAdditivityScheme(Scheme):
    def __init__(self, patterns=[], pretreatment_rules=[], remaps={},
                 other_descriptors=[], smiles_based_descriptors=[],
                 smarts_based_descriptors=[], include=[]):
        """Load group-additivity scheme from file-system `path` or builtin.

        Parameters
        ----------
        path : str
            Specify either the path to a file containing the data or a symbolic
            name of a builtin scheme to load (*e.g.* ``benson`` to load the
            standard Benson groups scheme.)

        Returns
        -------
        scheme : :class:`GroupScheme`
            Group scheme containing the loaded data.

        Raises
        ------
        pgradd.error.GroupSchemeError
            If `signature` is specified and does not match the signature of
            the loaded data.
        """
        self.patterns = patterns
        self.pretreatment_rules = pretreatment_rules
        self.remaps = remaps
        self.other_descriptors = other_descriptors
        self.smiles_based_descriptors = smiles_based_descriptors
        self.smarts_based_descriptors = smarts_based_descriptors
        for scheme in include:
            if isinstance(scheme, GroupAdditivityScheme):
                scheme_object = scheme
            else:
                scheme_object = self.Load(scheme['include'][0])
            self.patterns += scheme_object.patterns
            self.pretreatment_rules += scheme_object.pretreatment_rules
            self.remaps.update(scheme_object.remaps)
            self.other_descriptors += scheme_object.other_descriptors
            self.smiles_based_descriptors = scheme_object.\
                smiles_based_descriptors
            self.smarts_based_descriptors = scheme_object.\
                smarts_based_descriptors

    @classmethod
    def Load(cls, path):
        if os.sep not in path and '.' not in path and not os.path.exists(path):
            # [JTF] where's our data directory?
            base_path = os.path.join(get_data_dir(), path)
            # Load the scheme.yaml file in that directory:
            actual_path = os.path.join(base_path, 'scheme.yaml')
        else:
            actual_path = path
        abs_path = os.path.abspath(actual_path)
        fid = open(abs_path, 'r')
        scheme_data = yaml.load(fid, Loader=yaml.SafeLoader)
        fid.close()
        # change to molquery
        for i in range(0, len(scheme_data['patterns'])):
            scheme_data['patterns'][i]['connectivity'] = \
                Read(scheme_data['patterns'][i]['connectivity'])

        patterns = scheme_data['patterns']
        pretreatment_rules = []
        remaps = {}
        other_descriptors = []
        smiles_based_descriptors = []
        smarts_based_descriptors = []
        if 'pretreatment_rules' in scheme_data:
            pretreatment_rules = scheme_data['pretreatment_rules']
        if 'remaps' in scheme_data:
            remaps = scheme_data['remaps']
        if 'other_descriptors' in scheme_data:
            for i in range(0, len(scheme_data['other_descriptors'])):
                scheme_data['other_descriptors'][i]['connectivity'] = \
                    Read(scheme_data['other_descriptors'][i]['connectivity'])
            other_descriptors = scheme_data['other_descriptors']
        if 'smiles_based_descriptors' in scheme_data:
            for i in range(0, len(scheme_data['smiles_based_descriptors'])):
                scheme_data['smiles_based_descriptors'][i]['smarts'] = \
                    Chem.MolFromSmarts(scheme_data['smiles_based_descriptors']
                                       [i]['smarts'])
            smiles_based_descriptors = scheme_data['smiles_based_descriptors']
        if 'smarts_based_descriptors' in scheme_data:
            for i in range(0, len(scheme_data['smarts_based_descriptors'])):
                scheme_data['smarts_based_descriptors'][i]['smarts'] = \
                    Chem.MolFromSmarts(scheme_data['smarts_based_descriptors']
                                       [i]['smarts'])
            smarts_based_descriptors = scheme_data['smarts_based_descriptors']
        return cls(patterns, pretreatment_rules, remaps, other_descriptors,
                   smiles_based_descriptors, smarts_based_descriptors)

    def GetDescriptors(self, mol, debug=0):
        if isinstance(mol, Chem.Mol):

            mol = Chem.AddHs(mol)
            Chem.Kekulize(mol)
        elif isinstance(mol, str):
            clean_mol = Chem.MolFromSmiles(mol)
            # sanitize false is necessary to read adsorbed hydrogen species.
            # (e.g. [Pt][H])
            mol.replace('{', '')
            mol.replace('}', '')
            # mol = Chem.MolFromSmiles(mol,sanitize=False) # Stereochemistry
            # get's erased. :/
            mol = Chem.MolFromSmiles(mol)
            if mol is None:
                print('Smiles could not be loaded.')
                raise Exception
            sanitize_except_aromatization(mol)
            mol = Chem.AddHs(mol)
            Chem.Kekulize(mol)
        # pretreatment
        # TODO
        # reaction_query = Read(self.pretreatment_rules[0][0])
        # mol = reaction_query.ModifyMolInPlace(mol).
        # Change unspecified bond to ZERO, or weak bonds.
        for bond in mol.GetBonds():
            if bond.GetBondType().__str__() == 'UNSPECIFIED':
                bond.SetBondType(Chem.BondType.ZERO)
        # aromatize C6 ring for benson
        _aromatization_Benson(mol)
        # assign groups
        self._AssignCenterPattern(mol, debug)
        groups = self._AssignGroup(mol)
        descriptors = self._AssignDescriptor(mol, clean_mol)
        all_descriptors = groups.copy()
        all_descriptors.update(descriptors)
        return all_descriptors

    def _AssignCenterPattern(self, mol, debug=0):

        for pattern in self.patterns:

            matches = pattern['connectivity'].GetQueryMatches(mol)
            # only the first atom is the center atom
            matches = set([match[0] for match in matches])
            for match in matches:
                atom = mol.GetAtomWithIdx(match)
                if atom.HasProp('Group_Center_Name'):
                    s = 'Group center overwritten:' + \
                        atom.GetProp('Group_Center_Name') + ' to ' + \
                        pattern['center_name']
                    s += '(' + pattern['connectivity'].__str__() + ')'
                    raise PatternMatchError(s, atom)
                else:
                    atom.SetProp('Group_Center_Name', pattern['center_name'])
                    atom.SetProp('Group_Periph_Name', pattern['periph_name'])
            if debug:
                s = '\nPattern: ' + pattern['center_name'] +\
                    ',' + pattern['periph_name']
                s += '\nMatches: ' + str(matches)
                print(s)
        # exception spitted if any atoms have no groups assigned
        for atom in mol.GetAtoms():
            if not atom.HasProp('Group_Center_Name'):
                s = '\nSub group not assigned:\n' + Chem.MolToSmiles(mol) +\
                    ' Atom symbol: ' + atom.GetSymbol() + \
                    ' #radicalE: ' + atom.GetNumRadicalElectrons().__str__() +\
                    '\nConnectivity:'
                for bond in atom.GetBonds():
                    if bond.GetBeginAtomIdx() == atom.GetIdx():
                        s += '\n' + bond.GetEndAtom().GetSymbol() + ' ' + \
                            bond.GetBondType().__str__()
                    else:
                        s += '\n' + bond.GetBeginAtom().GetSymbol() + ' ' + \
                            bond.GetBondType().__str__()
                s += '\n Occured '
                raise PatternMatchError(s, atom)

    def _AssignGroup(self, mol):
        groups = defaultdict(int)
        for atom in mol.GetAtoms():
            csg = atom.GetProp('Group_Center_Name')
            if csg != 'none':
                psgs = list()
                for neighboratom in atom.GetNeighbors():
                    psg = neighboratom.GetProp('Group_Periph_Name')
                    if psg != 'none':
                        psgs.append(psg)
                group = Group(self, csg, psgs)
                groups[group.name] += 1
                atom.SetProp('Group_name', group.name)
            else:
                atom.SetProp('Group_name', 'none')
        # remapping
        if hasattr(self, 'remaps'):
            for atom in mol.GetAtoms():
                if atom.GetProp('Group_name') in self.remaps:
                    atom.SetProp('Group_name',
                                 self.remaps[atom.GetProp('Group_name')][0][1])
            for group in list(groups.keys()):
                if group in self.remaps:
                    n = groups.pop(group)
                    for remap in self.remaps[group]:
                        nn = n*remap[0]
                        groups[remap[1]] += nn
        return groups

    def _AssignDescriptor(self, mol, clean_mol):
        descriptors = defaultdict(int)
        for descriptor in self.other_descriptors:
            matches = descriptor['connectivity'].GetQueryMatches(mol)
            matches = set([tuple(set(match)) for match in matches])
            if matches:
                descriptors[descriptor['name']] += len(matches)
        for descriptor in self.smiles_based_descriptors:
            matches = clean_mol.GetSubstructMatches(descriptor['smiles'],
                                                    useChirality=descriptor
                                                    ['useChirality'])
            matches = set([tuple(set(match)) for match in matches])
            if matches:
                descriptors[descriptor['name']] += len(matches)
        for descriptor in self.smarts_based_descriptors:
            matches = mol.GetSubstructMatches(descriptor['smarts'],
                                              useChirality=descriptor
                                              ['useChirality'])
            matches = set([tuple(set(match)) for match in matches])
            if matches:
                descriptors[descriptor['name']] += len(matches)
        # remaps
        if hasattr(self, 'remaps'):
            for descriptor in list(descriptors.keys()):
                if descriptor in self.remaps:
                    n = descriptors.pop(descriptor)
                    for remap in self.remaps[descriptor]:
                        nn = n*remap[0]
                        descriptors[remap[1]] += nn
        return descriptors


def sanitize_except_aromatization(mol):
    Chem.SanitizeMol(mol, sanitizeOps=Chem.rdmolops.SanitizeFlags.
                     SANITIZE_ADJUSTHS)
    Chem.SanitizeMol(mol, sanitizeOps=Chem.rdmolops.SanitizeFlags.
                     SANITIZE_CLEANUP)
    Chem.SanitizeMol(mol, sanitizeOps=Chem.rdmolops.SanitizeFlags.
                     SANITIZE_CLEANUPCHIRALITY)
    Chem.SanitizeMol(mol, sanitizeOps=Chem.rdmolops.SanitizeFlags.
                     SANITIZE_FINDRADICALS)
    Chem.SanitizeMol(mol, sanitizeOps=Chem.rdmolops.SanitizeFlags.
                     SANITIZE_KEKULIZE)
    Chem.SanitizeMol(mol, sanitizeOps=Chem.rdmolops.SanitizeFlags.
                     SANITIZE_PROPERTIES)
    Chem.SanitizeMol(mol, sanitizeOps=Chem.rdmolops.SanitizeFlags.
                     SANITIZE_SETCONJUGATION)
    Chem.SanitizeMol(mol, sanitizeOps=Chem.rdmolops.SanitizeFlags.
                     SANITIZE_SETHYBRIDIZATION)
    Chem.SanitizeMol(mol, sanitizeOps=Chem.rdmolops.SanitizeFlags.
                     SANITIZE_SYMMRINGS)


def _aromatization_Benson(mol):
    # Benson's aromatic bond criteria:
    # C6 ring
    # No oxygen atoms
    # No triple bond (radical is okay)
    # Evenly spaced 3 double bond
    rings = Chem.GetSymmSSSR(mol)
    for i in range(0, len(rings)):
        # ring size check
        if len(rings[i]) != 6:
            continue
        # atom check
        if mol.GetAtomWithIdx(rings[i][0]).GetSymbol() != 'C':
            continue
        if mol.GetAtomWithIdx(rings[i][1]).GetSymbol() != 'C':
            continue
        if mol.GetAtomWithIdx(rings[i][2]).GetSymbol() != 'C':
            continue
        if mol.GetAtomWithIdx(rings[i][3]).GetSymbol() != 'C':
            continue
        if mol.GetAtomWithIdx(rings[i][4]).GetSymbol() != 'C':
            continue
        if mol.GetAtomWithIdx(rings[i][5]).GetSymbol() != 'C':
            continue

        # bond check
        if mol.GetBondBetweenAtoms(rings[i][0],
                                   rings[i][1]).\
                GetBondType().__str__() ==\
                'SINGLE':
            if mol.GetBondBetweenAtoms(rings[i][1],
                                       rings[i][2]).\
                    GetBondType().__str__() ==\
                    'DOUBLE':
                if mol.GetBondBetweenAtoms(rings[i][2],
                                           rings[i][3]).\
                        GetBondType().__str__() ==\
                        'SINGLE':
                    if mol.GetBondBetweenAtoms(rings[i][3],
                                               rings[i][4]).\
                            GetBondType().__str__() ==\
                            'DOUBLE':
                        if mol.GetBondBetweenAtoms(rings[i][4],
                                                   rings[i][5]).\
                                GetBondType().__str__() ==\
                                'SINGLE':
                            if mol.GetBondBetweenAtoms(rings[i][5],
                                                       rings[i][0]).\
                                    GetBondType().__str__() ==\
                                    'DOUBLE':
                                pass
                            else:
                                continue
                        else:
                            continue
                    else:
                        continue
                else:
                    continue
            else:
                continue
        elif mol.GetBondBetweenAtoms(rings[i][0],
                                     rings[i][1]).\
                GetBondType().__str__() ==\
                'DOUBLE':
            if mol.GetBondBetweenAtoms(rings[i][1],
                                       rings[i][2]).\
                    GetBondType().__str__() ==\
                    'SINGLE':
                if mol.GetBondBetweenAtoms(rings[i][2],
                                           rings[i][3]).\
                        GetBondType().__str__() ==\
                        'DOUBLE':
                    if mol.GetBondBetweenAtoms(rings[i][3],
                                               rings[i][4]).\
                            GetBondType().__str__() ==\
                            'SINGLE':
                        if mol.GetBondBetweenAtoms(rings[i][4],
                                                   rings[i][5]).\
                                GetBondType().__str__() ==\
                                'DOUBLE':
                            if mol.GetBondBetweenAtoms(rings[i][5],
                                                       rings[i][0]).\
                                    GetBondType().__str__() ==\
                                    'SINGLE':
                                pass
                            else:
                                continue
                        else:
                            continue
                    else:
                        continue
                else:
                    continue
            else:
                continue
        else:
            continue
        # if it survived all these continue, then this molecule is aromatic
        mol.GetAtomWithIdx(rings[i][0]).SetIsAromatic(True)
        mol.GetAtomWithIdx(rings[i][1]).SetIsAromatic(True)
        mol.GetAtomWithIdx(rings[i][2]).SetIsAromatic(True)
        mol.GetAtomWithIdx(rings[i][3]).SetIsAromatic(True)
        mol.GetAtomWithIdx(rings[i][4]).SetIsAromatic(True)
        mol.GetAtomWithIdx(rings[i][5]).SetIsAromatic(True)
        mol.GetBondBetweenAtoms(rings[i][0], rings[i][1]).SetIsAromatic(True)
        mol.GetBondBetweenAtoms(rings[i][1], rings[i][2]).SetIsAromatic(True)
        mol.GetBondBetweenAtoms(rings[i][2], rings[i][3]).SetIsAromatic(True)
        mol.GetBondBetweenAtoms(rings[i][3], rings[i][4]).SetIsAromatic(True)
        mol.GetBondBetweenAtoms(rings[i][4], rings[i][5]).SetIsAromatic(True)
        mol.GetBondBetweenAtoms(rings[i][5], rings[i][0]).SetIsAromatic(True)
        mol.GetBondBetweenAtoms(rings[i][0], rings[i][1]).SetIsConjugated(True)
        mol.GetBondBetweenAtoms(rings[i][1], rings[i][2]).SetIsConjugated(True)
        mol.GetBondBetweenAtoms(rings[i][2], rings[i][3]).SetIsConjugated(True)
        mol.GetBondBetweenAtoms(rings[i][3], rings[i][4]).SetIsConjugated(True)
        mol.GetBondBetweenAtoms(rings[i][4], rings[i][5]).SetIsConjugated(True)
        mol.GetBondBetweenAtoms(rings[i][5], rings[i][0]).SetIsConjugated(True)
        mol.GetBondBetweenAtoms(rings[i][0],
                                rings[i][1]).\
            SetBondType(Chem.rdchem.BondType.AROMATIC)
        mol.GetBondBetweenAtoms(rings[i][1],
                                rings[i][2]).\
            SetBondType(Chem.rdchem.BondType.AROMATIC)
        mol.GetBondBetweenAtoms(rings[i][2],
                                rings[i][3]).\
            SetBondType(Chem.rdchem.BondType.AROMATIC)
        mol.GetBondBetweenAtoms(rings[i][3],
                                rings[i][4]).\
            SetBondType(Chem.rdchem.BondType.AROMATIC)
        mol.GetBondBetweenAtoms(rings[i][4],
                                rings[i][5]).\
            SetBondType(Chem.rdchem.BondType.AROMATIC)
        mol.GetBondBetweenAtoms(rings[i][5],
                                rings[i][0]).\
            SetBondType(Chem.rdchem.BondType.AROMATIC)



"""
=========================================
Defenition to draw RDKIT mol object (:mod:`pgradd.DrawMol`)
=========================================

Coverts a rdkit mol object to a svg image and display.

"""

from rdkit import Chem
from rdkit.Chem import rdDepictor
from rdkit.Chem.Draw import rdMolDraw2D
from IPython.display import SVG, display


# http://rdkit.blogspot.com/2015/02/new-drawing-code.html
def moltosvg(mol, highlight=[], molSize=(400, 400), kekulize=True):
    mc = Chem.Mol(mol.ToBinary())
    if kekulize:
        try:
            Chem.Kekulize(mc)
        except Exception:
            mc = Chem.Mol(mol.ToBinary())
    if not mc.GetNumConformers():
        rdDepictor.Compute2DCoords(mc)

    drawer = rdMolDraw2D.MolDraw2DSVG(molSize[0], molSize[1])

    # Atom Label
    opts = drawer.drawOptions()
    # Atom name and index
    for i in range(mol.GetNumAtoms()):
        opts.atomLabels[i] = mol.GetAtomWithIdx(i).GetSymbol()+str(i)
    # radicals and charges
    for atom in mol.GetAtoms():
        nr = atom.GetNumRadicalElectrons()
        nc = atom.GetFormalCharge()
        if nr > 0:
            string = atom.GetSymbol() + ':'*divmod(nr, 2)[0] +\
                '.'*divmod(nr, 2)[1]
            opts.atomLabels[atom.GetIdx()] += string
        elif nc == 1:
            string = atom.GetSymbol() + '+'
            opts.atomLabels[atom.GetIdx()] += string
        elif nc > 1:
            string = atom.GetSymbol() + '+' + str(nc)
            opts.atomLabels[atom.GetIdx()] += string
        elif nc == -1:
            string = atom.GetSymbol() + '-'
            opts.atomLabels[atom.GetIdx()] += string
        elif nc < -1:
            string = atom.GetSymbol() + '-' + str(nc)
            opts.atomLabels[atom.GetIdx()] += string

    # highlight
    if highlight:
        drawer.DrawMolecule(mc, highlightAtoms=highlight)
    else:
        drawer.DrawMolecule(mc)

    drawer.FinishDrawing()
    svg = drawer.GetDrawingText()
    # It seems that the svg renderer used doesn't quite hit the spec.
    # Here are some fixes to make it work in the notebook, although I think
    # the underlying issue needs to be resolved at the generation step
    svg.replace('svg:', '')
    display(SVG(svg))



from .lib_interface import YAMLTaggedValue, parse
from .schema import SchemaRepository
# from .common import *


__all__ = ['Error', 'load', 'make_loader', 'make_object_loader',
           'register_class', 'register_list_type', 'register_type', 'parse']


class Error(Exception):
    """Exception raised upon YAML input/output related error."""


def load(data, context=None, name=None, tag=None, loader=None):
    """Construct object from abstract tree obtained using `parse()`."""
    if context is None:
        context = {}
    context['repo'] = _repository

    if tag is not None:
        if loader is not None:
            raise Error("Specification of keyword argument 'tag' conflicts "
                        "with specification of 'loader' in yaml_io.load()")
        return _repository.load_tagged(name, data, context, tag)
    elif loader is not None:
        if isinstance(data, YAMLTaggedValue):
            raise Error("Object %r with tag %r cannot be loaded with "
                        "loader %r in yaml_io.load().", loader)
        return loader(name, data, context)
    elif not isinstance(data, YAMLTaggedValue):
        raise Error("Cannot load untagged object %r; no tag or loader "
                    "specified to yaml_io.load()." % data)
    else:
        return _repository.load_tagged(name, data.value, context, data.tag)


_repository = SchemaRepository()

make_loader = _repository.make_loader
make_object_loader = _repository.make_object_loader
register_class = _repository.register_class
register_list_type = _repository.register_list_type
register_type = _repository.register_type



import numpy as np
from scipy.interpolate import InterpolatedUnivariateSpline
from scipy.integrate import quad as integrate
from .. import yaml_io
from .base import ThermochemBase
from ..Units import eval_qty

# from ..Consts import GAS_CONSTANT as R
import pmutt as pmutt
R = pmutt.constants.R(units='J/mol/K')
# R is now sourced from pmutt.constants instead of Consts


class ThermochemRawData(ThermochemBase):
    """
    Implement a thermochemical property correlation from raw data.

    Evaluated quantities are interpolated using a B-spline as discussed in
    :ref:`correlations documentation <correlations>`.
    """
    def __init__(self, ND_H_ref, ND_S_ref, Ts, ND_Cps,
                 T_ref=pmutt.constants.T0(units='K'), range=None):
        """
        Initialize a thermochemical property correlation from raw data.

        Parameters
        ----------
        ND_H_ref : float
            Non-dimensional standard heat of formation |eq_ND_H_ref|
        ND_S_ref : float
            Non-dimensional standard state reference entropy |eq_ND_S_ref|
        Ts : float array
            Temperatures at which `ND_Cps` are evaluated.
        ND_Cps : float
            Non-dimensional standard state heat capacities |eq_ND_Cp_T|
            evaluated at each temperature in `Ts`.
        T_ref : float, optional
            Reference temperature |eq_Tref| for `ND_H_ref` and `ND_S_ref`
            (default: room temperature according to pmutt, likely 298.15K).
        range : tuple(float, float), optional
            ``(lb, ub) = range`` where lb and ub are respectively the lower and
             uppers bounds of temperatures [K] for which the correlation is
             valid.    If specified, this range must contain T_ref and all data
             points in ND_Cp.
        """
        (self.Ts, self.ND_Cps) = list(zip(*sorted(
            zip(Ts, ND_Cps), key=lambda T_ND_Cps: T_ND_Cps[0])))
        self.min_T = Ts[0]
        self.max_T = Ts[-1]

        if range is None:
            range = (self.min_T, self.max_T)
        else:
            if self.min_T < range[0] or self.max_T > range[1]:
                raise ValueError(
                    'Heat capacity data points %g or %g lie outside of range'
                    ' [%g,%g].' % (self.min_T, self.max_T, range[0], range[1]))
        if T_ref < range[0] or T_ref > range[1]:
            raise ValueError(
                'T_ref=%g is outside the valid correlation range [%g,%g].'
                % (T_ref, range[0], range[1]))

        ThermochemBase.__init__(self, range)

        self.min_ND_Cp = self.ND_Cps[0]
        self.max_ND_Cp = self.ND_Cps[-1]

        self.ND_H_ref = ND_H_ref
        self.ND_S_ref = ND_S_ref
        self.T_ref = T_ref

        N = len(self.Ts)
        if N == 1:
            self.spline = ConstantSpline(self.ND_Cps[0])
        else:
            self.spline = InterpolatedUnivariateSpline(
                self.Ts, self.ND_Cps, k=(3 if N > 3 else N - 1))

    def get_CpoR(self, T):
        """Return non-dimensional standard state heat capacity |eq_ND_Cp_T|."""
        self.check_range(T)
        if not np.isscalar(T):
            return self._get_CpoR_ar(T)

        if T < self.min_T:
            return self.min_ND_Cp
        if T > self.max_T:
            return self.max_ND_Cp

        # Work-around for SciPy bug (?):
        # return self.spline(T)
        return float(self.spline(T))

    def _get_CpoR_ar(self, T):
        ND_Cp = np.empty(T.shape)
        T_below = T < self.min_T
        T_above = T > self.max_T
        T_middle = np.logical_not(np.logical_or(T_below, T_above))

        ND_Cp[T_below] = self.min_ND_Cp
        ND_Cp[T_above] = self.max_ND_Cp
        ND_Cp[T_middle] = self.spline(T[T_middle])

        return ND_Cp

    def get_SoR(self, T):
        """Return non-dimensional standard state entropy |eq_ND_S_T|."""
        self.check_range(T)
        T_a = self.T_ref
        T_b = T
        min_T = self.min_T
        max_T = self.max_T

        ND_S = self.ND_S_ref

        if T_a <= min_T:
            if T_b <= min_T:
                return ND_S + self.min_ND_Cp*np.log(T_b/T_a)
            ND_S += self.min_ND_Cp*np.log(min_T/T_a)
            T_a = min_T
        elif T_b <= min_T:
            ND_S += self.min_ND_Cp*np.log(T_b/min_T)
            T_b = min_T

        if T_a >= max_T:
            if T_b >= max_T:
                return ND_S + self.max_ND_Cp*np.log(T_b/T_a)
            ND_S += self.max_ND_Cp*np.log(max_T/T_a)
            T_a = max_T
        elif T_b >= max_T:
            ND_S += self.max_ND_Cp*np.log(T_b/max_T)
            T_b = max_T

        # The easiest, albeit not necessarily the best thing to do here is to
        # use numerical integration, so that's what we do.
        return ND_S + integrate(lambda t: self.spline(t)/t, T_a, T_b)[0]

    def get_HoRT(self, T):
        """Return non-dimensional standard heat of formation |eq_ND_H_T|."""
        self.check_range(T)
        T_a = self.T_ref
        T_b = T
        min_T = self.min_T
        max_T = self.max_T

        # This value represents the accumulated H/R (has temperature units).
        rH = self.ND_H_ref*T_a

        if T_a <= min_T:
            if T_b <= min_T:
                return (rH + self.min_ND_Cp*(T_b - T_a))/T
            rH += self.min_ND_Cp*(min_T - T_a)
            T_a = min_T
        elif T_b <= min_T:
            rH += self.min_ND_Cp*(T_b - min_T)
            T_b = min_T

        if T_a >= max_T:
            if T_b >= max_T:
                return rH + self.max_ND_Cp*(T_b - T_a)/T
            rH += self.max_ND_Cp*(max_T - T_a)
            T_a = max_T
        elif T_b >= max_T:
            rH += self.max_ND_Cp*(T_b - max_T)
            T_b = max_T

        return (rH + self.spline.integral(T_a, T_b))/T

    @classmethod
    def yaml_construct(cls, params, context):
        if 'T_ref' in params:
            T_ref = params['T_ref']
        else:
            #T_ref = eval_qty('298.15 K')  #fixed from eval_qty(298.15, 'K')
            T_ref = pmutt.constants.T0(units='K')
            #replaced getting room temp (298K) from eval_qty to pmutt.constants
        if 'ND_H_ref' in params:
            ND_H_ref = params['ND_H_ref']
        else:
            ND_H_ref = params['H_ref']/(R*T_ref)
        if 'ND_S_ref' in params:
            ND_S_ref = params['ND_S_ref']
        else:
            ND_S_ref = params['S_ref']/R

        if 'ND_Cp_data' in params:
            T_data, ND_Cp_data = list(zip(*params['ND_Cp_data']))
            Ts = np.array([T.in_units('K') for T in T_data])
            ND_Cps = np.array(ND_Cp_data)
        else:
            T_data, Cp_data = list(zip(*params['Cp_data']))
            Ts = np.array([T.in_units('K') for T in T_data])
            ND_Cps = np.array(
                [Cp for Cp in Cp_data])/R

        range = params.get('range')
        if range is not None:
            range = range[0].in_units('K'), range[1].in_units('K')
        else:
            range = Ts.min(), Ts.max()

        return cls(ND_H_ref, ND_S_ref, Ts, ND_Cps, T_ref.in_units('K'), range)

    _yaml_schema = """
range:
  type: tuple
  item_types: [{type: qty, kind: temperature},
               {type: qty, kind: temperature}]
  optional: True
  desc: range of valid temperatures

T_ref:
  type: qty
  kind: temperature
  default: 298.15 K
  desc: reference temperature for reference enthalpy and entropy

ND_Cp_data:
  type: list
  item_type:
     type: tuple
     item_types:
       - type: qty
         kind: temperature
       - type: float
  alts: Cp_data
  desc: set of (T, Cp(T)/R) data point pairs, where R is the gas constant

ND_H_ref:
  type: float
  alts: H_ref
  desc: H_ref/(R*T_ref) where R is the gas constant

ND_S_ref:
  type: float
  alts: ND_S_ref
  desc: S_ref/R where R is the gas constant

Cp_data:
  type: list
  item_type:
     type: tuple
     item_types:
       - type: qty
         kind: temperature
       - type: qty
         kind: molar heat capacity
  alts: ND_Cp_data
  desc: set of (T, Cp(T)) data point pairs

H_ref:
  type: qty
  kind: molar enthalpy
  alts: ND_H_ref
  desc: reference enthalpy

S_ref:
  type: qty
  kind: molar entropy
  alts: ND_S_ref
  desc: reference entropy
"""


yaml_io.register_class('ThermochemRawData',
                       yaml_io.parse(ThermochemRawData._yaml_schema),
                       ThermochemRawData)


class ConstantSpline(object):
    # This class emulates the interface to UnivariateSpline in the case of a
    # single data point (no interpolation).
    def __init__(self, ND_Cp):
        self.ND_Cp = ND_Cp

    def __call__(self, Ts):
        return self.ND_Cp*np.ones_like(Ts)

    def integral(self, T_a, T_b):
        return self.ND_Cp*(T_b - T_a)


__all__ = ['ThermochemRawData']



#
#
# setup.py
#
# Installation script to get setuptools to install pgradd into
# a Python environment.
#

import sys
import setuptools

# Import the lengthy rich-text README as the package's long
# description:
with open('README.rst', 'r') as fh:
	long_description = fh.read()

setuptools_info = {
	'name': 'pgradd',
	'version': '2.9.6',
	'author': 'Vlachos Research Group',
	'author_email': 'vlachos@udel.edu',
	'description': 'Python package implements the Group Additivity (GA) method for estimating thermodynamic properties',
	'long_description': long_description,
	'zip_safe': True,
	'url': 'https://github.com/VlachosGroup/PythonGroupAdditivity',
	'packages': setuptools.find_packages(),
	'include_package_data': True,
	'exclude_package_data': {
		'': [ 'README.rst', 'docs', 'example', 'tests', 'PKG-INFO', 'LICENSE.md' ]
	    },
	'install_requires': [
		'pmutt>=1.3.2',
		'scipy>=1.9.3',
		'numpy>=1.21.5',
		'IPython>=7.31.0',
		'PyYAML>=6.0',
		'python>=3.3'
	    ],
	'dependency_links': [
		'git+https://github.com/rdkit/rdkit/',
	    ],
	'classifiers': [
		"Programming Language :: Python :: 3",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent",
		"Intended Audience :: Science/Research",
		"Topic :: Scientific/Engineering :: Chemistry",
	    ],
    }

if sys.version_info[0] >= 3:
	#
	# Augment for Python 3 setuptools:
	#
	setuptools_info['long_description_content_type'] = 'text/x-rst'

setuptools.setup(**setuptools_info)



from .db import units_db
from .qty import Quantity, FundamentalUnits, eval_qty


# Define base SI units.
base_SI_units = [
  # (<base name>, <multiple>, <primitive name>)
  ('m', 1.0, 'm'),
  ('g', 0.001, 'kg'),
  ('s', 1.0, 's'),
  ('A', 1.0, 'A'),
  ('K', 1.0, 'K'),
  ('mol', 1.0, 'mol'),
  ('cd', 1.0, 'cd')
]
for base, mult, prim in base_SI_units:
    units_db.add(base, Quantity(mult, FundamentalUnits.new(prim)))

derived_SI_units = [
    ('N', 'kg m/s^2'), ('Pa', 'N/m^2'), ('J', 'N m'), ('W', 'J/s'),
    ('C', 'A s'), ('V', 'J/C'), ('F', 'C/V'), ('Ohm', 'V/A'),
    # etc.  (may define others later)
]
for name, val in derived_SI_units:
    units_db.add(name, eval_qty(val))

other_units = [
    # Includes non-SI metric units and English units.
    # Count:
    ('molecule', '1/(6.02214179*10^23) mol'),
    # Length:
    ('in', '2.54 cm'),
    ('ft', '12 in'),
    # Volume:
    ('L', '100 cm^3'),
    # Time:
    ('min', '60 s'),
    ('h', '60 min'),
    # Mass:
    ('u', '1.660538921*10^-27 kg'),
    ('lb', '0.45359237 kg'),
    ('t', '1000 kg'),
    # Force:
    ('dyn', '10^-5 N'),
    ('lbf', '4.44822162 N'),
    # Pressure:
    ('bar', '100.0 kPa'),
    ('atm', '101.325 kPa'),
    ('torr', '1.0/760 atm'),
    ('psi', 'lbf/in^2'),
    # Energy:
    ('cal', '4.184 J'),
    ('erg', 'dyn cm'),
    ('BTU', '1.05435026444 kJ'),
    ('eV', '1.602176487*10^-19 C V'),
    # Power:
    ('hp', '33000 ft lbf/min'),
    # Viscosity:
    ('P', 'g/(cm s)'),
    ('St', 'cm^2/s'),
]
for name, val in other_units:
    units_db.add(name, eval_qty(val))



import os
from warnings import warn
from collections.abc import Mapping
from .. import yaml_io
import numpy as np

from .. Error import GroupMissingDataError
from . Group import Group, Descriptor
from . Scheme import GroupAdditivityScheme
from . DataDir import get_data_dir


class GroupLibrary(Mapping):
    """Represent library of contributing properties organized by group.

    The set of properties that may be represented in a :class:`GroupLibrary` is
    extensible.  Contributing properties are represented by *property sets*
    organized by group.  See the manual for a list of available property sets.

    .. note::
        Because the *property set* system is extensible, the module within
        which a particular *property set* is defined and registered must be
        imported before loading a group library that contains data for that
        type of *property set*.

    To estimate properties, call :meth:`GroupLibrary.estimate()` with the
    *property set* `name` and  set of groups contained in the chemical
    structure of interest.  The properties estimate will be returned as an
    object whose type depends on the particular *property set*.

    To determine which groups are present in a particular chemical structure,
    use :meth:`GroupLibrary.match_groups()`.

    Data in multiple group libraries can be combined so long as the groups
    they contain are defined within compatible schemes.  See
    :meth:`GroupLibrary.update()`.

    """
    _property_set_estimator_types = {}
    _property_set_group_yaml_types = {}

    @classmethod
    def register_property_set_type(cls, name, group_yaml_type, estimator_type):
        """(class method) Register new property set type.

        Parameters
        ----------
        name : str
            Name of new property set type.
        group_yaml_type : str
            Name of property set type in the YAML type namespace.
        estimator_type : class
            The provided class is instantiated when an estimate is to be made
            for a particular set of groups.  The constructor should accept the
            following parameters:

                library : :class:`GroupLibrary`
                    library which is to be used to estimate these properties.
                groups : mapping
                    Map from :class:`Group` to int or float specifying counts
                    of each group in the chemical structure.
        """
        if name in cls._property_set_group_yaml_types:
            raise KeyError('Property set %r already registered.' % name)
        cls._property_set_group_yaml_types[name] = group_yaml_type
        cls._property_set_estimator_types[name] = estimator_type

    def __init__(self, scheme, contents={}, uq_contents={}, path=None):
        """Initialize library of contributing properties organized by group.

        Parameters
        ----------
        scheme : :class:`GroupScheme`
            Specify group-additivity scheme to use.
        contents : mapping or list
            Define initial contents of the library either as mapping or list of
            (`key`, `value`) pairs.  See the last paragraph of the class
            documentation for information on the format.

        Other Parameters
        ----------------
        path : str
            File-system path library was loaded from or should be saved to by
            default.
        """
        self.scheme = scheme
        self.path = path
        if isinstance(contents, Mapping):
            contents = list(contents.items())
        self.contents = dict((group, property_sets)
                             for (group, property_sets) in contents)
        self.uq_contents = uq_contents

    def GetDescriptors(self, mol):
        """Determine groups appearing in chemical structure `chem`.

        Parameters
        ----------
        mol : :class:`rdkit.mol`
            Specify chemical structure to match groups for.
        manual_descriptors : mapping, optional
            Specify value(s)/degree(s) of influence of additional descriptors
            to include.

        Returns
        -------
        groups : mapping
            Map from :class:`Group` to int or float identifying groups and
            their number of occurence in the structure.
        """
        self.name = mol
        return self.scheme.GetDescriptors(mol)

    def Estimate(self, groups, property_set_name):
        """Estimate set of properties for chemical.

        Parameters
        ----------
        groups : mapping (dictionary)
            Map from :class:`Group` to int or float specifying counts of each
            group in the chemical structure.
        property_set_name : str
            Name of property set to estimate.

        Returns
        -------
        estimated_properties : (varies)
            The estimated properties, an object whose type depends on the
            particular property set.
        """
        if property_set_name not in self._property_set_estimator_types:
            raise KeyError('Invalid property_set name: %r' % property_set_name)
        # Verify groups present.
        missing_groups = [group for group in groups
                          if property_set_name not in self[group]]
        if missing_groups:
            raise GroupMissingDataError(missing_groups, property_set_name)
        estimator_type = self._property_set_estimator_types[property_set_name]
        return estimator_type(self, groups)

    def __contains__(self, group):
        """Test if this library contains contributing properties for `group`.

        Parameters
        ----------
        group : :class:`Group`
            Group whose membership is being tested.

        Returns
        -------
        result : bool
            True if this library has properties for `group`.
        """
        return group in self.contents

    def __iter__(self):
        """Return iterator over all groups with property data in this library.
        """
        return iter(self.contents)

    def __len__(self):
        """Return number of groups with properties in this library."""
        return len(self.contents)

    def __getitem__(self, group):
        """Return contributing properties sets for `group`.

        If no properties exist for `group`, then return ``{}`` instead of
        raising an exception.

        Parameters
        ----------
        group : :class:`Group`
            Identify group whose property sets are to be retrieved.

        Returns
        -------
        property_sets : dict
            Sets of contributing properties for `group`.
        """
        return self.contents.get(group, {})

    @classmethod
    def Load(cls, path):
        """(class method) Load group-additivity library from file-system
        `path` or builtin.

        Parameters
        ----------
        path : str
            Specify either the path to a file containing the data or a symbolic
            name of a builtin library to load (*e.g.* ``gas_benson`` to load
            gas phase Benson groups.)

        Returns
        -------
        lib : :class:`GroupLibrary`
            Group library containing the loaded data.
        """
        if os.sep not in path and '.' not in path and not os.path.exists(path):
            # [JTF] where's our data directory?
            base_path = os.path.join(get_data_dir(), path)
            # We want to load the library.yaml in that directory:
            path = os.path.join(base_path, 'library.yaml')
        else:
            # The base path is the directory containing whatever file/directory
            # is referenced by path:
            base_path = os.path.dirname(path)

        # Load the scheme.yaml from the selected data directory:
        scheme = GroupAdditivityScheme.Load(os.path.join(base_path,
                                                         'scheme.yaml'))

        # Use that scheme to load the rest of the library:
        return cls._do_load(path, base_path, scheme)

    @classmethod
    def _Load(cls, path, scheme):
        if os.sep not in path and '.' not in path and not os.path.exists(path):
            # [JTF] where's our data directory?
            base_path = os.path.join(get_data_dir(), path)
            # We want to load the library.yaml in that directory:
            path = os.path.join(base_path, 'library.yaml')
        else:
            # The base path is the directory containing whatever file/directory
            # is referenced by path:
            base_path = os.path.dirname(path)

        # Use the scheme passed to us to load the rest of the library:
        return cls._do_load(path, base_path, scheme)

    @classmethod
    def _do_load(cls, path, base_path, scheme):
        # Read data from file.
        context = {'base_path': base_path}
        with open(path) as f:
            lib_data = yaml_io.load(
                yaml_io.parse(f.read()), context, loader=cls._yaml_loader)

        context['units'] = lib_data.units

        group_properties = lib_data.groups
        other_descriptor_properties = lib_data.other_descriptors
        UQ = lib_data.UQ

        if cls._property_set_group_yaml_types:
            # Prepare property_sets loader.
            property_sets_loader = yaml_io.make_object_loader(yaml_io.parse(
                '\n'.join(('%r:\n    type: %r\n    optional: true'
                          % (str(name),
                             str(cls._property_set_group_yaml_types[name])))
                          for name in cls._property_set_group_yaml_types)))
            # Read all properties.
            lib_contents = {}
            for name in group_properties:
                group = Group.parse(scheme, name)
                if group in lib_contents:
                    raise KeyError('Multiple definitions of group %s' % group)
                property_sets = yaml_io.load(
                    group_properties[name], context,
                    loader=property_sets_loader)
                lib_contents[group] = property_sets
            for name in other_descriptor_properties:
                descriptor = Descriptor(scheme, name)
                if descriptor in lib_contents:
                    raise KeyError('Multiple definitions of descriptor %s' %
                                   descriptor)
                property_sets = yaml_io.load(
                    other_descriptor_properties[name], context,
                    loader=property_sets_loader)
                lib_contents[descriptor] = property_sets

            # Read UQ data
            uq_contents = {}
            if UQ:
                uq_contents['RMSE'] = yaml_io.load(
                    UQ['RMSE'], context,
                    loader=property_sets_loader)
                uq_contents['descriptors'] = UQ['InvCovMat']['groups']
                uq_contents['mat'] = np.array(UQ['InvCovMat']['mat'])
                uq_contents['dof'] = UQ['DOF']
        else:
            # No property sets defined.
            warn('GroupLibrary.load(): No property sets defined.')
            lib_contents = {}
            uq_contents = {}

        new_lib = cls(scheme, lib_contents, uq_contents, path=path)
        # Update with included content.
        for include_path in lib_data.include:
            new_lib.Update(cls._Load(os.path.join(base_path,
                                                  include_path), scheme))
        return new_lib

    def Update(self, lib, overwrite=False):
        """Add complete contents of `lib` into this library.

        Parameters
        ----------
        lib : :class:`GroupLibrary`
            Library to import from.
        overwrite : bool
            If True, then existing data may be overwritten by data from `lib`.
        """
        for (group, other_property_sets) in list(lib.items()):
            if group not in self.contents:
                self.contents[group] = {}
            property_sets = self.contents[group]

            for name in other_property_sets:
                if name not in property_sets:
                    property_sets[name] = other_property_sets[name].copy()
                else:
                    property_sets[name].update(
                        other_property_sets[name], overwrite)
        # UQ stuff can only be loaded once
        if self.uq_contents and lib.uq_contents:
            raise ValueError('More than one uncertainty quantification',
                             'information provided')
        if not self.uq_contents:
            self.uq_contents = lib.uq_contents
    _yaml_loader = yaml_io.make_object_loader(yaml_io.parse("""
units:
    type: mapping
    default: {}

include:
    type: list
    item_type: string
    default: []

groups:
    type: mapping
    default: {}

other_descriptors:
    type: mapping
    default: {}

UQ:
    type: mapping
    default: {}
"""))



r"""
=====================================
RING notation for chemical structures
=====================================
This is a custom written abstract syntax tree parser for RING. Implementation
is based on the '<http://research.cems.umn.edu/bhan/#software>'.
Rangarajan, S., Kaminski, T., Van Wyk, E., Bhan A., and Daoutidis, P.,
"Language-oriented rule-based reaction network generation and analysis:
    Algorithms of RING", Computers and Chemical Engineering 64 (2014) 124,
    10.1016/j.compchemeng.2014.02.007
Rangarajan, S., Bhan, A., and Daoutidis, P.,
"Language-oriented rule-based reaction network generation and analysis:
    Descrpition of RING", Computers and Chemical Engineering 45 (2012) 114,
    10.1016/j.compchemeng.2012.06.008
"""

from .Reader import Read

__all__ = ['Read']



from .. ThermoChem import ThermochemBase, ThermochemIncomplete
from .. import yaml_io
import numpy as np
from ..GroupAdd.Library import GroupLibrary
from rdkit import Chem
from pmutt import constants as c

__all__ = ['ThermochemGroup', 'ThermochemGroupAdditive']


class ThermochemGroup(ThermochemIncomplete):
    """
    Implement thermochemical property correlation for
    one group's contributions.

    See base class :class:`ThermochemIncomplete` for additional documentation.
    """


yaml_io.register_class('ThermochemGroup',
                       yaml_io.parse(ThermochemGroup._yaml_schema),
                       ThermochemGroup)


class ThermochemGroupAdditive(ThermochemBase):
    """
    Implement thermochemical property correlation using group-additivity.

    Properties are evaluated by taking a linear combination of contributing
    properties of the constituent groups.

    Normally this class is not instantiated directly; rather, instances are
    created and returned by :meth:`GroupLibrary.estimate()`.
    """

    def __init__(self, lib, groups):
        """
        Initialize thermochemical property correlation using group-additivity.

        Parameters
        ----------
        lib : :class:`GroupLibrary`
            Specify library from which to access group data.
        groups : mapping
            Map from :class:`Group` to int specifying counts of each group in
            the chemical structure.
        """
        self.name = lib.name
        self.correlations = []
        common_min = None
        common_max = None

        for group in groups:
            count = groups[group]
            correlation = lib[group]['thermochem']
            self.correlations.append((correlation, count))
            data_range = correlation.get_range()
            if data_range is not None:
                if common_min is None:
                    (common_min, common_max) = data_range
                else:
                    common_min = max(common_min, data_range[0])
                    common_max = min(common_max, data_range[1])
        if lib.uq_contents:
            self.RMSE = lib.uq_contents['RMSE'].thermochem
            xp = np.zeros((len(lib.uq_contents['descriptors']), 1))
            for group in groups:
                count = groups[group]
                i = lib.uq_contents['descriptors'].index(group)
                xp[i] = count
            self.Xp_invXX_Xp = np.dot(np.dot(np.transpose(xp),
                                             lib.uq_contents['mat']), xp)
            self.dof = lib.uq_contents['dof']
        if common_min is not None:
            ThermochemBase.__init__(self, range=(common_min, common_max))
        else:
            ThermochemBase.__init__(self, range=None)

    def get_CpoR(self, T):
        return sum((count*correlation.get_CpoR(T)
                    for (correlation, count) in self.correlations))
    get_CpoR.__doc__ = ThermochemBase.get_CpoR.__doc__

    def get_HoRT(self, T):
        return sum((count*correlation.get_HoRT(T)
                    for (correlation, count) in self.correlations))
    get_HoRT.__doc__ = ThermochemBase.get_HoRT.__doc__

    def get_Selements(self):
        mol = Chem.rdmolops.AddHs(Chem.MolFromSmiles(self.name))
        atoms = mol.GetAtoms()
        S_ele = 0
        for atom in atoms:
            S_ele += c.S_elements[atom.GetAtomicNum()]
        return S_ele

    def get_SoR(self, T, S_elements=None):
        if not S_elements:
            S_ele = 0
        else:
            S_ele = self.get_Selements()
        return sum((count*correlation.get_SoR(T)
                    for (correlation, count) in self.correlations)) - S_ele
    get_SoR.__doc__ = ThermochemBase.get_SoR.__doc__

    def get_CpoR_SE(self, T):
        return float(np.sqrt(np.square(self.RMSE.get_CpoR(T)) *
                             self.Xp_invXX_Xp))

    def get_HoRT_SE(self, T):
        return float(np.sqrt(np.square(self.RMSE.get_HoRT(T)) *
                             self.Xp_invXX_Xp))

    def get_SoR_SE(self, T):
        return float(np.sqrt(np.square(self.RMSE.get_SoR(T)) *
                             self.Xp_invXX_Xp))


GroupLibrary.register_property_set_type(
    'thermochem', 'ThermochemGroup', ThermochemGroupAdditive)



# -*- coding: utf-8 -*-
"""
pmutt.test_constants
Tests for SMILES_to_thermo_examples file
Created on Tues Oct 23 2018
"""
import unittest
from pgradd.GroupAdd.Library import GroupLibrary
import pgradd.ThermoChem


class TestExamples(unittest.TestCase):

    def test_BensonGA(self):
        lib = GroupLibrary.Load('BensonGA')
        descriptors = lib.GetDescriptors('C1CO1')
        thermochem = lib.Estimate(descriptors, 'thermochem')
        GroupDict = {'C(C)(H)2(O)': 2, 'O(C)2': 1, 'Oxirane': 1}
        HoRT = thermochem.get_HoRT(298.15)
        self.assertAlmostEqual(HoRT, -21.09467743150278)
        self.assertEqual(descriptors, GroupDict)

    def test_Salciccioli(self):
        lib = GroupLibrary.Load('SalciccioliGA2012')
        descriptors = lib.GetDescriptors('C([Pt])C[Pt]')
        thermochem = lib.Estimate(descriptors, 'thermochem')
        GroupDict = {'C(C)(H)2(Pt)': 2, 'surface-ring strain': 0.217}
        H = thermochem.get_H(298.15, units='kcal/mol')
        self.assertAlmostEqual(H, -11.307743997749277)
        self.assertEqual(descriptors, GroupDict)

    def test_GU_Aq(self):
        lib = GroupLibrary.Load('GuSolventGA2017Aq')
        descriptors = lib.GetDescriptors('C(=O)([Pt])O')
        thermochem = lib.Estimate(descriptors, 'thermochem')
        GroupDict = {'CO(O)(Pt)+O(CO)(H)': 1.0}
        HoRT = thermochem.get_HoRT(500)
        self.assertAlmostEqual(HoRT, -109.86212002776878)
        self.assertEqual(descriptors, GroupDict)

    def test_Wittreich_Surface_ND(self):
        lib = GroupLibrary.Load('GRWSurface2018')
        descriptors = lib.GetDescriptors('[Pt]C([Pt])C([Pt])([Pt])C=O')
        thermochem = lib.Estimate(descriptors, 'thermochem')
        GroupDict = {'C(C)(H)(Pt)2': 1, 'C(C)(CO)(Pt)2': 1, 'CO(C)(H)': 1,
                     'CPt2CPt2': 1, 'CCPt2': 1, 'surface-ring strain': 0.392}
        HoRT = thermochem.get_HoRT(750)
        self.assertAlmostEqual(HoRT, -13.423119203382337)
        self.assertEqual(descriptors, GroupDict)

    def test_Wittreich_Surface_Dim(self):
        lib = GroupLibrary.Load('GRWSurface2018')
        descriptors = lib.GetDescriptors('[Pt]C([Pt])C([Pt])([Pt])C=O')
        thermochem = lib.Estimate(descriptors, 'thermochem')
        GroupDict = {'C(C)(H)(Pt)2': 1, 'C(C)(CO)(Pt)2': 1, 'CO(C)(H)': 1,
                     'CPt2CPt2': 1, 'CCPt2': 1, 'surface-ring strain': 0.392}
        H = thermochem.get_H(750, 'kcal/mol')
        self.assertAlmostEqual(H, -20.005853103142883)
        self.assertEqual(descriptors, GroupDict)

    def test_Wittreich_Aqueous_ND(self):
        lib = GroupLibrary.Load('GRWAqueous2018')
        descriptors = lib.GetDescriptors('C(=O)([Pt])O')
        thermochem = lib.Estimate(descriptors, 'thermochem')
        GroupDict = {'CO(O)(Pt)+O(CO)(H)': 1.0}
        HoRT = thermochem.get_HoRT(500)
        self.assertAlmostEqual(HoRT, -107.57909464133714)
        self.assertEqual(descriptors, GroupDict)

    def test_Wittreich_Aqueous_Dim(self):
        lib = GroupLibrary.Load('GRWAqueous2018')
        descriptors = lib.GetDescriptors('C(=O)([Pt])O')
        thermochem = lib.Estimate(descriptors, 'thermochem')
        GroupDict = {'CO(O)(Pt)+O(CO)(H)': 1.0}
        H = thermochem.get_H(500, 'kJ/mol')
        self.assertAlmostEqual(H, -447.23102885789655)
        self.assertEqual(descriptors, GroupDict)

    def test_Xie_Ru0001_Example1_ND(self):
        lib = GroupLibrary.Load('XieGA2022')
        descriptors = lib.GetDescriptors('[Ru]C([Ru])C([Ru])([Ru])C')
        thermochem = lib.Estimate(descriptors, 'thermochem')
        GroupDict = {'C(C)(H)(Ru)2': 1, 'C(C)2(Ru)2': 1, 'C(C)(H)3': 1,
                     'CRu2CRu2': 1}
        HoRT = thermochem.get_HoRT(500)
        self.assertAlmostEqual(HoRT, -35.040312149773726)
        self.assertEqual(descriptors, GroupDict)

    def test_Xie_Ru0001_Example1_Dim(self):
        lib = GroupLibrary.Load('XieGA2022')
        descriptors = lib.GetDescriptors('[Ru]C([Ru])C([Ru])([Ru])C')
        thermochem = lib.Estimate(descriptors, 'thermochem')
        GroupDict = {'C(C)(H)(Ru)2': 1, 'C(C)2(Ru)2': 1, 'C(C)(H)3': 1,
                     'CRu2CRu2': 1}
        H = thermochem.get_H(500, 'kJ/mol')
        self.assertAlmostEqual(H, -145.6706333743726)
        self.assertEqual(descriptors, GroupDict)

    def test_Xie_Ru0001_Example2_ND(self):
        lib = GroupLibrary.Load('XieGA2022')
        descriptors = lib.GetDescriptors('CCC')
        thermochem = lib.Estimate(descriptors, 'thermochem')
        GroupDict = {'C(C)(H)3': 2, 'C(C)2(H)2': 1}
        HoRT = thermochem.get_HoRT(500)
        self.assertAlmostEqual(HoRT, -41.49969417868688)
        self.assertEqual(descriptors, GroupDict)

    def test_Xie_Ru0001_Example2_Dim(self):
        lib = GroupLibrary.Load('XieGA2022')
        descriptors = lib.GetDescriptors('CCC')
        thermochem = lib.Estimate(descriptors, 'thermochem')
        GroupDict = {'C(C)(H)3': 2, 'C(C)2(H)2': 1}
        H = thermochem.get_H(500, 'kJ/mol')
        self.assertAlmostEqual(H, -172.52376948049303)
        self.assertEqual(descriptors, GroupDict)

    def test_Selements_Example_Dim(self):
        lib = GroupLibrary.Load('BensonGA')
        descriptors = lib.GetDescriptors('CCCCCC')
        thermochem = lib.Estimate(descriptors, 'thermochem')
        GroupDict = {'C(C)(H)3': 2, 'C(C)2(H)2': 4}
        G = thermochem.get_G(T=298.15, units='kJ/mol', S_elements=True)
        self.assertAlmostEqual(G, -7.732446702038452)
        self.assertEqual(descriptors, GroupDict)


if __name__ == '__main__':
    unittest.main()
